from __future__ import annotations

import logging
import sys
from dataclasses import dataclass
from enum import Enum
from io import StringIO
from logging import Logger
from typing import Optional


# ---------------------------------------------------------

class LoggerFactory:
    @classmethod
    def get_logger(cls, name : str,
                   log_fpath : Optional[str] = None,
                   include_timestamp : bool = True,
                   include_location : bool = False,
                   include_logger_name : bool = False,
                   threshold : int = logging.INFO,
                   use_stdout : bool = False) -> Logger:
        if LoggerOverseer.logger_exists(name=name):
            return logging.getLogger(name)

        logger = logging.getLogger(name=name)
        logger.setLevel(threshold)
        formatting = Formatting(print_timestamp=include_timestamp,
                                print_location=include_location,
                                print_logger_name=include_logger_name)

        stream = sys.stdout if use_stdout else sys.stderr
        console_handler = logging.StreamHandler(stream)
        console_formatter = Formatter(log_target=LogTarget.CONSOLE, formatting=formatting)
        console_handler.setFormatter(console_formatter)
        logger.addHandler(console_handler)

        if log_fpath:
            file_handler = logging.FileHandler(log_fpath)
            file_formatter = Formatter(log_target=LogTarget.FILE, formatting=formatting)
            file_handler.setFormatter(file_formatter)
            logger.addHandler(file_handler)

        return logger


class LoggerOverseer:
    @staticmethod
    def logger_exists(name : str) -> bool:
        logger_names = [name for name in logging.root.manager.loggerDict]
        return name in logger_names


    @staticmethod
    def force_identification():
        logging.basicConfig(level=logging.DEBUG)
        root_logger = logging.getLogger()
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        for handler in root_logger.handlers:
            handler.setFormatter(formatter)

    @staticmethod
    def show_loggers():
        logger_names = [name for name in logging.root.manager.loggerDict]
        print("-> Currently running Loggers:")
        for name in logger_names:
            print(f"- {name}")
        return logger_names

    @staticmethod
    def redirect(logger : logging.Logger, new_stream : StringIO):
        for h in logger.handlers:
            logger.removeHandler(hdlr=h)
        handler = logging.StreamHandler(new_stream)
        logger.addHandler(handler)

class Formatter(logging.Formatter):
    custom_file_name = 'custom_file_name'
    custom_line_no = 'custom_lineno'
    colors: dict = {
        logging.DEBUG: '\033[20m',
        logging.INFO: '\033[20m',
        logging.WARNING: '\033[93m',
        logging.ERROR: '\033[91m',
        logging.CRITICAL: '\x1b[31;1m'  # Bold Red
    }

    def __init__(self, log_target : LogTarget, formatting : Formatting):
        self.formatting : Formatting = formatting
        self.log_target : LogTarget = log_target
        super().__init__()


    def format(self, record):
        log_fmt = "%(message)s"

        custom_time = self.formatTime(record, "%Y-%m-%d %H:%M:%S")
        conditional_timestamp = f"{custom_time}" if self.formatting.print_timestamp else ''
        conditional_name = f"{record.name}" if self.formatting.print_logger_name else ''

        extra_format = ''
        if self.formatting.print_timestamp:
            extra_format = conditional_timestamp

        if self.formatting.print_logger_name:
            conditional_space = ' ' if extra_format else ''
            extra_format = f'{extra_format}{conditional_space}{conditional_name}'

        if extra_format:
            log_fmt = f"[{extra_format}]: {log_fmt}"

        if self.formatting.print_location:
            log_fmt += f'\t\t| Location: File "{record.pathname}:{record.lineno}"'

        if self.log_target == LogTarget.CONSOLE:
            color_prefix = Formatter.colors.get(record.levelno, "")
            color_suffix = "\033[0m"
            log_fmt = color_prefix + log_fmt + color_suffix

        self._style._fmt = log_fmt
        return super().format(record)

class LogTarget(Enum):
    FILE = "FILE"
    CONSOLE = "CONSOLE"


@dataclass
class Formatting:
    print_timestamp : bool
    print_logger_name : bool
    print_location : bool




import dataclasses
from dataclasses import dataclass
from datetime import datetime, date, time
from decimal import Decimal
from enum import Enum
from types import NoneType
from typing import get_type_hints, get_origin, get_args, Union

import orjson
from PIL.Image import Image

from holytools.abstract.serialization.serializable import Serializable
from holytools.fileIO import ImageConverter

BasicSerializable = (bool | int | float | str | Serializable | Decimal | datetime | date | time | Enum | Image | None)

# -------------------------------------------

@dataclass
class JsonDataclass(Serializable):
    """Enables serialization of dataclasses with following attributes:
    - Basic serializable types as defined above:
        - Serialization: get_basic_entry()
        - Deserialization: make_basic()
    - Lists, tuples or dicts of basic serializable types"""
    def __init__(self, *args, **kwargs):
        _, __ = args, kwargs
        if not dataclasses.is_dataclass(self):
            raise TypeError(f'{self.__class__} must be a dataclass to be Jsonifyable')

    def to_str(self) -> str:
        defined_fields = set([f.name for f in dataclasses.fields(self) if f.init])
        json_dict = {}
        for attr, value in [(attr, value) for attr, value in self.__dict__.items() if attr in defined_fields]:
            if isinstance(value, list):
                entry = [self.get_basic_entry(x) for x in value]
            elif isinstance(value, tuple):
                entry = tuple([self.get_basic_entry(x) for x in value])
            elif isinstance(value, dict):
                key_list = [self.get_basic_entry(k) for k in value.keys()]
                value_list = [self.get_basic_entry(v) for v in value.values()]
                entry = (key_list, value_list)
            else:
                entry = self.get_basic_entry(obj=value)
            json_dict[attr] = entry
        
        return orjson.dumps(json_dict).decode("utf-8")

    @classmethod
    def from_str(cls, json_str: str):
        if not dataclasses.is_dataclass(cls):
            raise TypeError(f'{cls} is not a dataclass. from_json can only be used with dataclasses')

        json_dict = orjson.loads(json_str)
        type_hints = get_type_hints(cls)
        init_dict = {}
        for key, value in json_dict.items():
            dtype = type_hints.get(key)
            if TypeAnalzer.is_optional(dtype) and value is None:
                init_dict[key] = None
                continue
            if dtype is None:
                continue

            dtype = TypeAnalzer.strip_nonetype(dtype)
            origin = get_origin(dtype)
            if origin == list:
                item_type = TypeAnalzer.get_inner_types(dtype)[0]
                restored_value = [cls.make_basic(basic_cls=item_type, s=x) for x in value]
            elif origin == tuple:
                item_types = TypeAnalzer.get_inner_types(dtype)
                restored_value = tuple([cls.make_basic(basic_cls=item_type, s=s) for item_type, s in zip(item_types, value)])
            elif origin == dict or dtype == dict:
                key_type, value_type = TypeAnalzer.get_inner_types(dtype)
                key_list = [cls.make_basic(basic_cls=key_type, s=x) for x in value[0]]
                value_list = [cls.make_basic(basic_cls=value_type, s=x) for x in value[1]]
                restored_value = {key: value for key, value in zip(key_list, value_list)}
            else:
                restored_value = cls.make_basic(basic_cls=dtype, s=value)
            init_dict[key] = restored_value

        return cls(**init_dict)

    @staticmethod
    def get_basic_entry(obj) -> (str | int | NoneType):
        if not isinstance(obj, BasicSerializable):
            raise TypeError(f'Object {obj} is not a basic serializable type')

        if obj is None:
            entry = None
        elif isinstance(obj, Serializable):
            entry = obj.to_str()
        elif isinstance(obj, Enum):
            entry = obj.name
        elif isinstance(obj, float) and obj != obj:
            entry = 'nan'
        elif isinstance(obj, Image):
            entry = ImageConverter.to_base64_str(image=obj)
        elif isinstance(obj, bool):
             entry = bool(int(obj))
        else:
            entry = str(obj)
        return entry

    @staticmethod
    def make_basic(basic_cls, s : str):
        castable_classes = ['str', 'int', 'float', 'bool', 'Decimal', 'UUID', 'Path']
        converters = {
            datetime: datetime.fromisoformat,
            date : date.fromisoformat,
            time : time.fromisoformat,
        }

        if basic_cls in converters:
            instance = converters[basic_cls](s)
        elif basic_cls.__name__ in castable_classes:
            instance = basic_cls(s)
        elif issubclass(basic_cls, Enum):
            instance =  basic_cls[s]
        elif issubclass(basic_cls, Serializable):
            instance = basic_cls.from_str(s)
        elif basic_cls == Image:
            instance = ImageConverter.from_base64_str(s)
        elif basic_cls == NoneType:
            instance = None
        else:
            raise TypeError(f'Unsupported type {basic_cls}')
        return instance


class TypeAnalzer:
    @staticmethod
    def is_optional(dtype):
        origin = get_origin(dtype)
        if origin is Union:
            return NoneType in get_args(dtype)
        else:
            return False

    # noinspection DuplicatedCode
    @staticmethod
    def strip_nonetype(dtype : type) -> type:
        origin = get_origin(dtype)
        if origin is Union:
            types = get_args(dtype)
            not_none_types = [t for t in types if not t is NoneType]
            if len(not_none_types) == 1:
                core_type = not_none_types[0]
            else:
                raise ValueError(f'Union dtype {dtype} has more than one core dtype')
        else:
            core_type = dtype
        return core_type
    
    @staticmethod
    def get_inner_types(dtype : type) -> tuple:
        inner_dtypes = get_args(dtype)
        return inner_dtypes


if __name__ == "__main__":
    pass


import dill
from pickle import dumps, loads
from .serializable import Serializable

class Dillable(Serializable):
    def to_str(self) -> str:
        return dill.dumps(self).hex()

    @classmethod
    def from_str(cls, s: str):
        return dill.loads(bytes.fromhex(s))


class Picklable(Serializable):
    def to_str(self) -> str:
        return dumps(self).hex()

    @classmethod
    def from_str(cls, s: str):
        return loads(bytes.fromhex(s))


# Example usage:
if __name__ == "__main__":
    class Example(Dillable):
        def __init__(self, data):
            self.data = data

    original = Example(data="Sample data")
    test_dill_str = original.to_str()
    restored = Example.from_str(test_dill_str)

    print(f"Original: {original.data}")
    # print(f"Restored: {restored._execution_profiles}")



import psutil
import inspect
from psutil import Process, NoSuchProcess
from pympler import asizeof

# -------------------------------------------

class ProcessStatistics:
    def __init__(self, pid : int):
        self.pid : int = pid

        init_successful = False
        try:
            self.process : Process = psutil.Process(self.pid)
            init_successful = True
        except NoSuchProcess:
            pass
        if not init_successful:
            raise IOError(f"Process with PID {self.pid} does not exist.")

    def get_subprocesses(self) -> list[Process]:
        return self.process.children(recursive=True)

    def get_memory_usage_in_mb(self) -> float:
        memory_info = self.process.memory_info()
        return memory_info.rss / (1024 ** 2)

    def get_cpu_usage(self) -> float:
        process = psutil.Process(self.pid)
        return process.cpu_percent(interval=0.1)

    def get_subprocess_mem_usage(self) -> dict[str, int]:
        mem_usage = {}
        for child in self.get_subprocesses():
            try:
                mem_info = child.memory_info()
                mem_usage[child.name()] = mem_info.rss
            except psutil.NoSuchProcess:
                continue
        return mem_usage


class FunctionStatistics:
    @staticmethod
    def get_mem_usage() -> int:
        caller_frame = inspect.currentframe().f_back
        return asizeof.asizeof(caller_frame.f_locals)

    @staticmethod
    def get_variable_mem_usage() -> dict[str, int]:
        caller_frame = inspect.currentframe().f_back
        variables_memory_usage = {}
        for var_name, var_value in caller_frame.f_locals.items():
            variables_memory_usage[var_name] = asizeof.asizeof(var_value)
        return variables_memory_usage


if __name__ == "__main__":
    process_stats = ProcessStatistics(4028)
    print(f"Memory usage: {process_stats.get_memory_usage_in_mb()} MB")
    print(f"CPU usage: {process_stats.get_cpu_usage()} %")
    print(process_stats.get_subprocesses())
    print(process_stats.get_subprocess_mem_usage())


from pathlib import Path as PathWrapper
import tempfile, shutil
import os, stat
from typing import Optional


class FsysNode:
    def __init__(self, path : str):
        self._path_wrapper : PathWrapper = PathWrapper(path)

    def get_zip(self) -> bytes:
        with tempfile.TemporaryDirectory() as write_dir:
            zip_base_path = os.path.join(write_dir, 'zipfile')
            args_dir = {
                'base_name': zip_base_path,
                'format': 'zip',
            }
            if self._path_wrapper.is_file():
                args_dir['root_dir'] = self.get_parent_dirpath()
                args_dir['base_dir'] = self.get_name()

            if self._path_wrapper.is_dir():
                args_dir['root_dir'] = self.get_path()

            shutil.make_archive(**args_dir)
            with open(f'{zip_base_path}.zip', 'rb') as file:
                zip_bytes = file.read()

        return zip_bytes

    # -------------------------------------------
    # attributes

    def get_path(self) -> str:
        return str(self._path_wrapper.absolute())

    def get_parent_dirpath(self) -> str:
        return str(self._path_wrapper.parent.absolute())

    def get_name(self) -> str:
        return self._path_wrapper.name

    def is_hidden(self) -> bool:
        if os.name == 'posix':
            return os.path.basename(self.get_path()).startswith('.')
        elif os.name == 'nt':
            return bool(os.stat(self.get_path()).st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN)
        else:
            raise NotImplementedError(f'Unsupported OS: {os.name}, {FsysNode.is_hidden.__name__} is only supported '
                                      f'on Windows and Unix systems')

    def get_last_modified_epochtime(self) -> float:
        return os.path.getmtime(self.get_path())

    def get_size_in_MB(self) -> float:
        return os.path.getsize(self.get_path()) / (1024 * 1024)


class Directory(FsysNode):
    def __init__(self, path : str):
        super().__init__(path=path)
        if not self._path_wrapper.is_dir():
            raise FileNotFoundError(f'Path {path} is not a directory')

    def get_tree(self, include_root : bool = False) -> str:
        fpaths = self.get_all_subpaths(absolute=include_root)
        structure_dict = self.to_dict(fpaths)
        return self.dict_to_tree(fs_dict=structure_dict)

    def get_subfile_fpaths(self, absolute : bool = True) -> list[str]:
        subfile_paths = []
        for root, dirs, files in os.walk(self.get_path()):
            for file in files:
                fpath = os.path.join(root, file)
                subfile_paths.append(fpath)
        if not absolute:
            subfile_paths = [os.path.relpath(p, self.get_path()) for p in subfile_paths]
        return subfile_paths

    def get_all_subpaths(self, absolute : bool = True) -> list[str]:
        all_paths = []
        for root, dirs, files in os.walk(self.get_path()):
            for d in dirs:
                dirpath = os.path.join(root, d)
                all_paths.append(dirpath)
            for f in files:
                fpath = os.path.join(root, f)
                all_paths.append(fpath)
        if not absolute:
            all_paths = [os.path.relpath(p, self.get_path()) for p in all_paths]

        return all_paths

    @classmethod
    def dict_to_tree(cls, fs_dict: dict, indent: int = 0, parent_dirpath : str = '', max_children: int = 10) -> str:
        total_str = ''
        if not fs_dict:
            return total_str

        items = fs_dict.items()
        for n, (k, v) in enumerate(items):
            indentation = '\t' * indent

            if n == max_children:
                total_str += f'{indentation}... (Max folder elements displayed = {max_children})\n'
                break

            parent_name = os.path.basename(parent_dirpath)
            parent_indicator = f'-> {parent_name}/' if parent_name else ''

            fpath = os.path.join(parent_dirpath, k)
            is_file = os.path.isfile(fpath)
            symbol = f'🗎' if is_file else '🗀'
            conditional_backslash = '' if is_file else '/'
            total_str += (f'{indentation}{symbol} {k}{conditional_backslash} {parent_indicator}\n'
                          f'{cls.dict_to_tree(v, indent=indent + 1, parent_dirpath=fpath)}')
        if indent == 0:
            total_str = total_str.rstrip()
        return total_str

    @staticmethod
    def to_dict(fpaths: list[str]) -> dict:
        tree_dict = {}
        for path in fpaths:
            parts = path.split('/')
            node = tree_dict
            for p in parts:
                if not p in node:
                    node[p] = {}
                node = node[p]
        return tree_dict

class File(FsysNode):
    def __init__(self, path : str):
        super().__init__(path=path)
        if not self._path_wrapper.is_file():
            raise FileNotFoundError(f'Path {path} is not a file')

    def get_suffix(self) -> Optional[str]:
        parts = self.get_name().split('.')
        if len(parts) == 1:
            return None
        else:
            return parts[-1]


if __name__ == "__main__":
    directory = Directory(path='/home/daniel/misc/holytools', )
    print(directory.get_tree())


from __future__ import annotations
from __future__ import annotations

import linecache
import logging
import os
import time
import traceback
import unittest
from dataclasses import dataclass
from typing import Optional
from unittest import TestCase, TestResult

from holytools.devtools.testing.basetest import BaseTest

# ---------------------------------------------------------

class Result(TestResult):
    test_spaces = 50
    status_spaces = 10
    runtime_space = 10

    def __init__(self, logger : logging.Logger,
                 testsuite_name: str,
                 manual_mode : bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logger
        self.reports : list[Report] = []
        self.start_times : dict[str, float] = {}
        self.is_manual : bool = manual_mode
        self.testsuite_name = testsuite_name

    def startTestRun(self):
        super().startTestRun()
        self.log(self.get_header(f'  Test suite: \"{self.testsuite_name}\"  '))

    def startTest(self, case : BaseTest):
        super().startTest(case)
        self.on_case_start(case=case)

    def stopTestRun(self):
        self.log_summary()
        super().stopTestRun()

    # noinspection PyTypeChecker
    def addSuccess(self, case : BaseTest):
        super().addSuccess(case)
        self.on_case_finish(case, CaseStatus.SUCCESS)

    # noinspection PyTypeChecker
    def addError(self, case : BaseTest, err):
        super().addError(case, err)
        self.on_case_finish(case, CaseStatus.ERROR, err)

    # noinspection PyTypeChecker
    def addFailure(self, case : BaseTest, err):
        super().addFailure(case, err)
        self.on_case_finish(case, CaseStatus.FAIL, err)

    # noinspection
    def addSkip(self, case : BaseTest, reason):
        super().addSkip(case, reason)
        self.on_case_finish(case, CaseStatus.SKIPPED)

    # ---------------------------------------------------------
    # case reports

    def on_case_start(self, case : BaseTest):
        if self.is_manual:
            case.set_is_manual()
        self.log_test_start(casename=case.get_name())
        self.start_times[case.id()] = time.time()

    def on_case_finish(self, case : BaseTest, status: str, err : Optional[tuple] = None):
        report = Report(name=case.get_name(), status=status, runtime=self._get_runtime(case), err=err)
        self.log(msg=report.get_view(), level=report.get_log_level())
        self.reports.append(report)


    def _get_runtime(self, test : TestCase) -> Optional[float]:
        test_id = test.id()
        if test_id in self.start_times:
            time_in_sec =  time.time() - self.start_times[test_id]
            return round(time_in_sec, 3)
        else:
            self.log(msg=f'Couldnt find start time of test {test_id}. Current start_times : {self.start_times}',
                     level=logging.ERROR)

    # ---------------------------------------------------------
    # logging

    def log_summary(self):
        self.log(self.get_header(msg=' Summary ', seperator='-'))
        for case in self.reports:
            level = case.get_log_level()
            name_msg = f'{case.name[:self.test_spaces - 4]:<{self.test_spaces}}'
            status_msg = f'{case.status:<{self.status_spaces}}'
            runtime_str = f'{case.runtime}s'
            runtime_msg = f'{runtime_str:^{self.runtime_space}}'

            self.log(msg=f'{name_msg}{status_msg}{runtime_msg}', level=level)
        self.log(self.get_final_status())
        self.log(self.get_header(msg=''))


    def get_header(self, msg: str, seperator : str = '=') -> str:
        total_len = self.test_spaces + self.status_spaces
        total_len += self.runtime_space
        line_len = max(total_len- len(msg), 0)
        lines = seperator * int(line_len / 2.)
        return f'{lines}{msg}{lines}'


    def get_final_status(self) -> str:
        num_total = len(self.reports)
        unsuccessful_cases = [c for c in self.reports if c.status != CaseStatus.SUCCESS]
        num_unsuccessful = len(unsuccessful_cases)

        RED = '\033[91m'
        GREEN = '\033[92m'
        RESET = '\033[0m'
        CHECKMARK = '✓'
        CROSS = '❌'

        if num_unsuccessful == 0:
            final_status = f"{GREEN}\n{CHECKMARK} {num_total}/{num_total} tests ran successfully!{RESET}"
        else:
            final_status = f"{RED}\n{CROSS} {num_unsuccessful}/{num_total} tests had errors or failures!{RESET}"

        return final_status

    def log_test_start(self, casename : str):
        self.log(msg=f'------> {casename[:self.test_spaces]} ', level=logging.INFO)

    def log(self, msg : str, level : int = logging.INFO):
        self.logger.log(msg=msg, level=level)


@dataclass
class Report:
    runtime : float
    name : str
    status : str
    err : Optional[Exception] = None

    def get_log_level(self) -> int:
        status_to_logging : dict[str, int] = {
            CaseStatus.SUCCESS: logging.INFO,
            CaseStatus.ERROR: logging.CRITICAL,
            CaseStatus.FAIL: logging.ERROR,
            CaseStatus.SKIPPED : logging.INFO
        }
        return status_to_logging[self.status]


    def get_view(self) -> str:
        conditional_err_msg = f'\n{self._get_err_details(self.err)}' if self.err else ''
        return f'Status: {self.status}{conditional_err_msg}\n'


    @staticmethod
    def _get_err_details(err) -> str:
        err_class, err_instance, err_traceback = err
        tb_list = traceback.extract_tb(err_traceback)

        def is_relevant(tb):
            not_unittest = not os.path.dirname(unittest.__file__) in tb.filename
            not_custom_unittest = not os.path.dirname(__file__) in tb.filename
            return not_unittest and not_custom_unittest

        user_tb = [tb for tb in tb_list if is_relevant(tb)]

        result = ''
        relevant_tb = user_tb if not len(user_tb) == 0 else tb_list
        for frame in relevant_tb:
            file_path = frame.filename
            line_number = frame.lineno
            tb_str = (f'File "{file_path}", line {line_number}, in {frame.name}\n'
                      f'    {linecache.getline(file_path, line_number).strip()}')
            result += f'{err_class.__name__}: {err_instance}\n{tb_str}'
        return result


class CaseStatus:
    SUCCESS = 'SUCCESS'
    ERROR = 'ERROR'
    FAIL = 'FAIL'
    SKIPPED = 'SKIPPED'



import time
from collections.abc import Collection
from typing import Iterator

import progressbar
from progressbar import ProgressBar


# ---------------------------------------------------------

class TrackedInt:
    def __init__(self, start_value : int, finish_value : int):
        if start_value >= finish_value:
            raise ValueError(f'Start value {start_value} must be less than finish value {finish_value}')
        self._value : int = start_value
        self.progressbar = ProgressBar(min_value=start_value, max_value=finish_value)
        self.progressbar.update()

    def increment(self, to_add : int = 1):
        self._value += to_add

        if self.progressbar.finished():
            return

        if self._value >= self.progressbar.max_value:
            self.finish()
            return

        self.progressbar.update(value=self._value)

    def finish(self):
        self.progressbar.finish()

    def is_finished(self) -> bool:
        return self.progressbar.finished()

    def get_value(self) -> int:
        return self._value

    @classmethod
    def wrap_output(cls):
        progressbar.streams.wrap_stderr()
        progressbar.streams.wrap_stdout()

    # ------------------
    # int operators

    def __iadd__(self, other):
        if not isinstance(other, int):
            raise ValueError("Only integers can be added to a TrackedInt.")
        self.increment(to_add=other)
        return self

    def __add__(self, other):
        raise NotImplementedError(f'Can only add to tracked integer in place')

    def __eq__(self, other):
        if isinstance(other, TrackedInt):
            return self._value == other._value
        elif isinstance(other, int):
            return self._value == other
        return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, TrackedInt):
            return self._value < other._value
        elif isinstance(other, int):
            return self._value < other
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, TrackedInt):
            return self._value <= other._value
        elif isinstance(other, int):
            return self._value <= other
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, TrackedInt):
            return self._value > other._value
        elif isinstance(other, int):
            return self._value > other
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, TrackedInt):
            return self._value >= other._value
        elif isinstance(other, int):
            return self._value >= other
        return NotImplemented

    def as_range(self) -> Iterator[int]:
        while self._value < int(self.progressbar.max_value):
            yield self._value
            self.increment(to_add=1)

    def __str__(self):
        return str(self._value)




class TrackedCollection(Iterator):
    def __init__(self, iterable : Collection):
        self.inner : Iterator = iter(iterable)
        self.tracking_int : TrackedInt = TrackedInt(start_value=0, finish_value=len(iterable))

    def __next__(self) -> object:
        self.tracking_int.increment(to_add=1)
        return next(self.inner)


if __name__ == "__main__":
    this = TrackedInt(start_value=1, finish_value=10)
    for k in this.as_range():
        print(f'k = {k}')
        time.sleep(0.1)

    for j in TrackedCollection(range(10)):
        time.sleep(0.1)
        print(f'j = {j}')



from __future__ import annotations
import os
from abc import abstractmethod
from typing import Optional


class Serializable:
    @abstractmethod
    def to_str(self) -> str:
        pass

    @classmethod
    @abstractmethod
    def from_str(cls, s: str):
        pass

    def save(self, fpath : str, force_overwrite : bool = False) -> str:
        fpath = os.path.abspath(path=fpath)
        dirpath = os.path.dirname(fpath)
        os.makedirs(dirpath, exist_ok=True)
        full_name = os.path.basename(fpath)
        non_suffixname = full_name.split('.')[0]
        suffix = get_suffix(fpath)

        if os.path.isfile(fpath) and not force_overwrite:
            fpath = get_free_path(save_dirpath=dirpath, name=non_suffixname, suffix=suffix, start_index=1)
            print(f'Warning: File already exists at specified filepath. Saving to {fpath} instead.')

        with open(fpath, 'w') as f:
            f.write(self.to_str())

        return fpath

    @classmethod
    def load(cls, fpath : str):
        with open(fpath, 'r') as f:
            str_data = f.read()
        return cls.from_str(str_data)


def get_suffix(fpath) -> Optional[str]:
    parts = fpath.split('.')
    if len(parts) == 1:
        return None
    else:
        return parts[-1]


def get_free_path(save_dirpath : str, name : str, suffix : Optional[str] = None, start_index : int = 0) -> str:
    if suffix:
        if not suffix.startswith('.'):
            suffix = f'.{suffix}'

    def get_path(index : int = start_index):
        conditional_suffix = '' if suffix is None else f'{suffix}'
        conditional_index = f'_{index}' if not index is None else ''
        return os.path.join(save_dirpath, f'{name}{conditional_index}{conditional_suffix}')

    fpath = get_path()
    current_index = 0
    while os.path.isfile(path=fpath) or os.path.isdir(fpath):
        current_index += 1
        fpath = get_path(index=current_index)
    return fpath




from __future__ import annotations

import PIL.Image as ImgHandler
from PIL.Image import Image

from .fileio import FileIO


# ---------------------------------------------------------


class ImageFile(FileIO):
    def check_content_ok(self) -> bool:
        pass

    def read(self) -> Image:
        return ImgHandler.open(self.fpath)

    def write(self, image: Image):
        image.save(self.fpath)

    def view(self):
        with self.read() as image:
            image.show()



from __future__ import annotations

import base64
from enum import Enum
from io import BytesIO

import PIL.Image as ImgHandler
from PIL import Image


class ImageConverter:
    PNG_MODES = ['L', 'LA', 'RGB', 'RGBA', 'P']
    JPG_MODES = ['L', 'RGB', 'CMYK']

    @classmethod
    def convert_format(cls, image: Image, target_format : ImageFormat) -> Image:
        if not image.format:
            raise TypeError(f'Given image {image} has no format')
        if not image.format.lower() in ImageFormat.get_all_formats():
            raise TypeError(f'Given image {image} has unsupported format: {image.format}')
        if not target_format in ImageFormat.get_all_formats():
            raise TypeError(f'Conversion to format {target_format} is not supported')
        if not ImageConverter._is_valid(image=image):
            raise TypeError(f'Image mode {image.mode} is invalid for format {image.format}')

        new_format = target_format.value.upper()
        if new_format in ['JPG', 'JPEG'] and not image.mode in cls.JPG_MODES:
            image = ImageConverter.to_rgb(image)
        if new_format == 'PNG' and not image.mode in cls.PNG_MODES:
            image = ImageConverter.to_rgba(image)
        else:
            image = image

        return ImageConverter._reload_as_fmt(image=image, target_format=target_format.value)

    @staticmethod
    def to_rgb(img : Image):
        new_img = ImgHandler.new('RGB', img.size, (255, 255, 255))
        rgb_content = img.convert('RGB')
        new_img.paste(rgb_content, mask=img.split()[-1])

        return ImageConverter._reload_as_fmt(new_img, target_format=img.format)

    @staticmethod
    def to_rgba(img):
        new_img = img.convert('RGBA')
        return ImageConverter._reload_as_fmt(new_img, target_format=img.format)

    @classmethod
    def _is_valid(cls, image: Image) -> bool:
        if not image.format.lower() in ImageFormat.get_all_formats():
            raise TypeError(f'Image format {image.format} is not supported')

        if image.format.upper() == 'PNG':
            return image.mode in cls.PNG_MODES
        elif image.format.upper() in ['JPEG', 'JPG']:
            return image.mode in cls.JPG_MODES
        return False

    @staticmethod
    def _reload_as_fmt(image : Image, target_format : str):
        buffer = BytesIO()
        image.save(buffer, format=target_format)
        buffer.seek(0)
        return ImgHandler.open(buffer)

    # --------------------------------------------------------------
    # serialization

    @staticmethod
    def as_bytes(image: Image) -> bytes:
        if image.format is None:
            raise TypeError(f'Given image {image} has no format')

        buffer = BytesIO()
        image.save(buffer, format=image.format)
        img_bytes = buffer.getvalue()
        return img_bytes

    @staticmethod
    def from_bytes(img_bytes : bytes) -> Image:
        buffer = BytesIO(img_bytes)
        image = ImgHandler.open(buffer)
        return image

    @staticmethod
    def from_base64_str(s : str) -> Image:
        if "base64," in s:
            s = s.split("base64,")[-1]
        image_data = base64.b64decode(s)
        image = ImgHandler.open(BytesIO(image_data))
        return image

    @staticmethod
    def to_base64_str(image: Image) -> str:
        byte_content = ImageConverter.as_bytes(image=image)
        base64_content = base64.b64encode(byte_content).decode('utf-8')
        return base64_content


class ImageFormat(Enum):
    PNG = 'PNG'
    JPG = 'JPG'
    JPEG = 'JPEG'

    @classmethod
    def get_all_formats(cls) -> list[ImageFormat]:
        return [member for member in cls]

    def __eq__(self, other):
        if isinstance(other,str):
            return self.value.upper() == other.upper()
        elif isinstance(other, ImageFormat):
            return self.value == other.value
        return False

    def __str__(self):
        return self.value


from __future__ import annotations

import io
import linecache
import os
import re
import sys
import traceback
import types
from datetime import datetime
from typing import Callable
from typing import Literal

Color = Literal['BLACK', 'RED', 'GREEN', 'YELLOW', 'BLUE', 'MAGENTA', 'CYAN', 'WHITE']

# ---------------------------------------------------------


class LoggingTools:
    @staticmethod
    def add_color(msg : str, color : Color) -> str:
        color_code_map = {
            'BLACK': '\033[30m',
            'RED': '\033[31m',
            'GREEN': '\033[32m',
            'YELLOW': '\033[33m',
            'BLUE': '\033[34m',
            'MAGENTA': '\033[35m',
            'CYAN': '\033[36m',
            'WHITE': '\033[37m'
        }

        color_code = color_code_map[color]
        return f"{color_code}{msg}\033[0m"

    @staticmethod
    def get_limited_stacktrace(err: Exception, excluded_modules : list[types.ModuleType]) -> str:
        err_class, err_instance, err_traceback = err.__class__, err, err.__traceback__
        tb_list = traceback.extract_tb(err_traceback)
        def is_relevant(tb):
            in_exclued_module = any([os.path.dirname(module.__file__) in tb.filename for module in excluded_modules])
            return not in_exclued_module

        err_msg = ''
        relevant_tb = [tb for tb in tb_list if is_relevant(tb)]
        if relevant_tb:
            err_msg = "\nEncountered error during training routine. Relevant stacktrace:"
            for frame in relevant_tb:
                file_path = frame.filename
                line_number = frame.lineno
                tb_str = (f'File "{file_path}", line {line_number}, in {frame.name}\n'
                          f'    {linecache.getline(file_path, line_number).strip()}')
                err_msg += f'\n{err_class.__name__}: {err_instance}\n{tb_str}'

        return err_msg

    @staticmethod
    def get_timestamp(time_only : bool = False):
        if time_only:
            return datetime.now().strftime("%H:%M:%S")
        else:
            return datetime.now().strftime("%Y-%m-%d_%H%M")

    @staticmethod
    def mute(func : Callable) -> Callable:
        def muted_func(*args, **kwargs):
            stdout, stderr = sys.stdout, sys.stderr
            sys.stdout = open(os.devnull, 'w')
            sys.stderr = open(os.devnull, 'w')
            result = func(*args, **kwargs)
            sys.stdout, sys.stderr = stdout, stderr
            return result
        return muted_func

    @staticmethod
    def to_sci_notation(val : str | float | int) -> str:
        try:
            val = float(val)
            display_val = f'{val:.1e}'
        except:
            display_val = val
        return  display_val


class CaptureLogs:
    def __init__(self):
        self.copied_stream = StoredStream(stdout=sys.stdout, stderr=sys.stderr)

    def get_stored(self) -> str:
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        plain_value = self.copied_stream.get_value()
        result = ansi_escape.sub('', plain_value)
        return result

    def __enter__(self):
        sys.stdout, sys.stderr = self.copied_stream, self.copied_stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__


class StoredStream:
    def __init__(self, stdout, stderr):
        self.original_stdout = stdout
        self.original_stderr = stderr
        self.buffer = io.StringIO()

    def write(self, data):
        self.original_stdout.write(data)
        self.original_stdout.flush()
        self.buffer.write(data)

    def flush(self):
        self.original_stdout.flush()
        self.original_stderr.flush()
        self.buffer.flush()

    def get_value(self):
        return self.buffer.getvalue()


if __name__ == "__main__":
    pass


import base64


class BytesConverter:
    @staticmethod
    def to_base64(data : bytes) -> str:
        return base64.b64encode(data).decode(encoding='utf-8')

    @staticmethod
    def from_base64(data : str) -> bytes:
        return base64.b64decode(data)

    @staticmethod
    def from_hex(data : str) -> bytes:
        return bytes.fromhex(data)

    @staticmethod
    def to_hex(data : bytes) -> str:
        return data.hex()

    @staticmethod
    def decode(data : bytes) -> str:
        return data.decode()

    @staticmethod
    def encode(data : str) -> bytes:
        return data.encode()


import socket
from enum import Enum
import subprocess
import requests
from holytools.network.adapter import Adapter


class NetworkArea(Enum):
    LOCALHOST  = 'LOCALHOST'
    HOME = 'HOME'
    GLOBAL = 'WAN'


class IpProvider:
    @classmethod
    def get_localhost(cls) -> str:
        return cls.get_ip(area=NetworkArea.LOCALHOST)

    @classmethod
    def get_ip(cls, area: NetworkArea) -> str:
        if area == NetworkArea.LOCALHOST:
            return '127.0.0.1'
        elif area == NetworkArea.HOME:
            return cls.get_private_ip()
        elif area == NetworkArea.GLOBAL:
            raise PermissionError("Unable to retrieve Global IP automatically. Please check manually")
        else:
            raise ValueError(f"Invalid network area: {area.value}")

    @staticmethod
    def get_private_ip() -> str:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(0)
        try:
            random_ip = '10.254.254.254'
            s.connect((random_ip, 1))
            IP = s.getsockname()[0]
        except Exception:
            IP = '127.0.0.1'
        s.close()
        return IP

    @staticmethod
    def get_public_ip() -> str:
        err, public_ip = None, None
        try:
            response = requests.get('https://api.ipify.org')
            if response.status_code == 200:
                public_ip = response.text
            else:
                err = ConnectionError(f'Unable to retrieve public IP: {response.status_code}')
        except Exception as e:
            err = e
        if err:
            raise err
        return public_ip


    @staticmethod
    def get_ipconfig() -> str:
        command = "nmcli device show"
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output, error = process.communicate()

        if error:
            raise Exception(f"Error executing nmcli: {error.decode()}")

        ipconfig = ''
        sections = "".join(output.decode()).split("\n\n")

        for section in sections:
            if section.strip():
                new_adapter = Adapter.from_nmcli_output(section)
                ipconfig += f'\n\n{new_adapter}'

        return ipconfig

    @staticmethod
    def get_free_port() -> int:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(('', 0))
            socknum = s.getsockname()[1]
        return socknum


if __name__ == "__main__":
    print(IpProvider.get_ipconfig())



import os
from typing import Optional

from pathvalidate import sanitize_filename


class PathTools:
    @staticmethod
    def to_valid_filename(name : str) -> str:
        name = name.strip()
        name = name.replace(' ', '_')
        name = sanitize_filename(name)
    
        return name

    @staticmethod
    def increment_idx_until_free(initial_dirpath : str) -> str:
        idx = 0
        dirpath = initial_dirpath
        while os.path.isdir(dirpath):
            idx += 1
            dirpath = f'{initial_dirpath}_{idx}'
        return dirpath

    @staticmethod
    def prune_suffix(fpath : str) -> str:
        parts = fpath.split('.')
        if len(parts) > 1:
            return '.'.join(parts[:-1])
        else:
            return fpath

    @staticmethod
    def ensure_suffix(fpath: str, suffix : str) -> str:
        parts = fpath.split('.')
        if len(parts) < 2:
            fpath = f'{fpath}.{suffix}'
        elif parts[-1] != f'.{suffix}':
            fpath = f'{parts[0]}.{suffix}'
        return fpath

    @staticmethod
    def get_suffix(fpath: str) -> Optional[str]:
        parts = fpath.split('.')
        if len(parts) > 1:
            return parts[-1]
        else:
            return None


import os.path


class SegmentProvider:
    def __init__(self, fpath : str, delimiter : str):
        if not os.path.isfile(fpath):
            raise FileNotFoundError(f'File {fpath} not found')
        with open(fpath, 'r') as f:
            file_content : str = f.read()
        segments = file_content.split(delimiter)

        self.segement_map : dict[str, str] = {}
        for s in segments:
            segment_name = s.split('\n')[0]
            content_lines = s.split('\n')[1:]
            segment_content = '\n'.join(content_lines)
            self.segement_map[segment_name] = segment_content

    def retrieve(self, name : str):
        """Retrieves a prompt from the specified prompt file by name"""
        return self.segement_map[name]




from .fileio import FileIO

class PlaintextFile(FileIO):
    def read(self) -> str:
        with open(self.fpath, 'r') as f:
            text = f.read()
        return text

    def write(self, content: str):
        with open(self.fpath, 'w', encoding='utf-8') as file:
            file.write(content)

    def view(self):
        content = self.read()
        print(content)

    def check_content_ok(self):
        try:
            with open(self.fpath, 'rb') as file:
                bytes_content = file.read()
                bytes_content.decode('utf-8')
            return True
        except UnicodeDecodeError:
            raise TypeError(f'File {self.fpath} is not a valid utf-8 encoded text file')

    @classmethod
    def get_text(cls, fpath : str) -> str:
        with open(fpath, 'r') as f:
            text = f.read()
        return text



import builtins


class InputSimulator:
    def __init__(self, inputs : list[str]):
        def generator():
            for inp in inputs:
                yield inp

        self.gen = generator()
        self.original_input = input

    def substitute(self):
        builtins.input = self.simulated_input

    def restore_input(self):
        builtins.input = self.original_input

    def simulated_input(self, *args, **kwargs):
        _ = kwargs
        try:
            response = next(self.gen)
        except:
            response = self.original_input()
        combined = f'{args[0]} {response}' if args else response
        print(combined)
        return response

if __name__ == "__main__":
    a = InputSimulator([f'An ice cream please','No, thank you'])
    a.substitute()

    item = input('What do you want?')
    r = input(f'Anything else?')

    print(f'done')


import inspect
from typing import Any, Callable, get_type_hints
from dataclasses import dataclass
from typing import Union

class BoundFunction:
    __self__: object


@dataclass
class Argument:
    name: str
    dtype: type

    def set_default_val(self, val: object):
        setattr(self, 'default_val', val)

    def has_default_val(self):
        return hasattr(self, 'default_val')

    def get_default_val(self) -> Any:
        if not self.has_default_val():
            raise AttributeError(f"Argument '{self.name}' has no default value.")
        return getattr(self, 'default_val')



class ModuleInspector:
    @classmethod
    def get_methods(cls, obj : Union[object, type],
                    include_inherited: bool = True,
                    include_private = False,
                    include_magic_methods : bool = False) -> list[Callable]:

        def mthd_ok(mthd_name : str) -> bool:
            if cls.is_magical(mthd_name):
                return include_magic_methods
            elif cls.is_private(mthd_name):
                return include_private
            else:
                return True

        attrs = dir(obj)
        if not include_inherited:
            obj_cls = obj if isinstance(obj, type) else obj.__class__
            parent_attrs = []
            for p in obj_cls.__bases__:
                parent_attrs += dir(p)
            attrs = [name for name in attrs if not name in parent_attrs]
        attr_values = [getattr(obj, name) for name in attrs]

        targeted_methods = []
        for attr_name, value in zip(attrs, attr_values):
            if callable(value) and mthd_ok(attr_name):
                targeted_methods.append(value)

        return targeted_methods

    @staticmethod
    def get_args(func: Union[Callable, BoundFunction], exclude_self : bool = True) -> list[Argument]:
        spec = inspect.getfullargspec(func)
        type_hints = get_type_hints(func)
        if not spec.args:
            return []

        start_index = 1 if exclude_self and spec.args[0] in ['self', 'cls'] else 0
        relevant_arg_names = spec.args[start_index:]
        defaults_mapping = ModuleInspector._get_defaults_mapping(spec=spec)

        def create_arg(name : str):
            dtype = type_hints.get(name)
            is_bound = inspect.ismethod(func)
            if name == 'self' and is_bound:
                dtype = func.__self__.__class__
            if name == 'cls' and is_bound:
                dtype = func.__self__
            if not dtype:
                raise ValueError(f"Type hint for argument '{name}' is missing.")
            argument = Argument(name=name, dtype=dtype)
            if name in defaults_mapping:
                argument.set_default_val(defaults_mapping[name])
            return argument

        return [create_arg(name=name) for name in relevant_arg_names]

    @staticmethod
    def _get_defaults_mapping(spec):
        defaults = spec.defaults or ()
        reversed_args = spec.args[::-1]
        reversed_defaults = defaults[::-1]

        zipped = zip(reversed_args, reversed_defaults)
        return dict(zipped)

    @staticmethod
    def is_magical(name : str):
        return name.startswith('__') and name.endswith('__')

    @staticmethod
    def is_private(name : str):
        return name.startswith('_')



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

import requests
from requests import Response


# -----------------------------------------------------------

class Method(Enum):
    GET = 'GET'
    POST = 'POST'


@dataclass
class Endpoint:
    ip : str
    port : int
    path : str

    @classmethod
    def make_localhost(cls, port : int, path : str):
        return cls(ip='127.0.0.1', port=port, path=path)

    def get_url(self, protocol : str) -> str:
        socket_addr = f'{protocol}://{self.ip}:{self.port}'
        return f'{socket_addr}{self.path}'

    def get(self, secure : bool = True) -> Response:
        protocol = 'https' if secure else 'http'
        url = self.get_url(protocol=protocol)
        print(f'Making get request to {url}')
        return requests.get(url)

    def post(self, msg : str, secure : bool = True) -> Response:
        protocol = 'https' if secure else 'http'
        url = self.get_url(protocol=protocol)
        return requests.post(url=url, data=msg)



import logging

from holytools.logging import LoggerFactory


class Timber:
    def __init__(self):
        self.logger = LoggerFactory.get_logger(name=self.__class__.__name__)

    def log(self, msg : str, level : int = logging.INFO):
        self.logger.log(level=level, msg=msg)

    def warning(self, msg : str, *args, **kwargs):
        kwargs['level'] = logging.WARNING
        self.logger.log(msg=msg, *args, **kwargs)

    def error(self, msg : str, *args, **kwargs):
        kwargs['level'] = logging.ERROR
        self.logger.log(msg=msg, *args, **kwargs)

    def critical(self, msg : str, *args, **kwargs):
        kwargs['level'] = logging.CRITICAL
        self.logger.log(msg=msg, *args, **kwargs)

    def info(self, msg : str, *args, **kwargs):
        kwargs['level'] = logging.INFO
        self.logger.log(msg=msg, *args, **kwargs)






class MessageFormatter:
    @staticmethod
    def get_boxed(text: str, headline: str = "") -> str:
        text = text.replace('\t', ' ' * 4)
        lines = text.split("\n")

        max_length = max([len(line) for line in lines] + [len(headline)+2])
        border = "+" + "-" * (max_length + 2) + "+"
        if headline:
            headline = f' {headline} '
            top_line = f"+{headline.center(max_length + 2, '-')}+"
        else:
            top_line =  border
        bottom_line = border

        boxed_lines = [top_line] + [f"| {line.ljust(max_length)} |" for line in lines] + [bottom_line]
        return "\n".join(boxed_lines) + '\n'

    @staticmethod
    def multi_section_box(texts: list[str], headlines: list[str]) -> str:
        combined_text = ''.join(texts)
        text = combined_text.replace('\t', ' ' * 4)
        lines = text.split("\n")

        max_length = max([len(line) for line in lines] + [len(h)+2 for h in headlines])
        border = "+" + "-" * (max_length + 2) + "+"

        lines = []
        for t, h in zip(texts, headlines):
            h = f' {h} '
            top_line = f"+{h.center(max_length + 2, '-')}+"
            lines += [top_line] + [f"| {line.ljust(max_length)} |" for line in t.split("\n")]
        lines += [border]

        return "\n".join(lines) + '\n'

    @staticmethod
    def get_boxed_train(messages: list) -> str:
        if len(messages) == 0:
            raise ValueError("No messages to format")

        for msg in messages:
            if '\n' in msg:
                raise ValueError(f"Message contains newline character: {msg}")
        wagons = [MessageFormatter.get_boxed(headline="", text=msg) for msg in messages]

        total_str = ''
        no_lines = 3
        wagon_sep = 5

        pulled_cabins = [w.split('\n') for w in wagons[:-1]]
        conductor_lines = wagons[-1].split('\n')

        for j in range(no_lines):
            connect_symbol = '-' if j == 1 else ' '
            connector = f'{connect_symbol * wagon_sep}'
            for lines in pulled_cabins:
                total_str += f'{lines[j]}{connector}'
            total_str += f'{conductor_lines[j]}\n'

        total_str = total_str.strip()

        return total_str



from __future__ import annotations
from typing import Optional

class Adapter:

    @classmethod
    def from_nmcli_output(cls,section: str) -> Adapter:
        the_adapter = Adapter()
        for line in section.split('\n'):
            if ':' not in line:
                continue
            key, value = line.split(':', 1)
            value = value.strip()

            if key == "GENERAL.DEVICE":
                the_adapter.name = value
            elif key == "GENERAL.TYPE":
                the_adapter.adapter_type = value
            elif key == "IP4.ADDRESS[1]":
                the_adapter.ipv4_address = value.split('/')[0]
                the_adapter.subnet_mask = value.split('/')[1] if '/' in value else None
            elif key == "IP4.GATEWAY":
                the_adapter.default_gateway = value
            elif key == "IP6.ADDRESS[1]":
                the_adapter.ipv6_address = value.split('/')[0]
            elif key == "IP4.DOMAIN[1]":
                the_adapter.dns_suffix = value

        return the_adapter

    def __init__(
        self,
        technical_name: Optional[str] = None,
        adapter_type: Optional[str] = None,
        dns_suffix: Optional[str] = None,
        ipv4_address: Optional[str] = None,
        ipv6_address: Optional[str] = None,
        subnet_mask: Optional[str] = None,
        default_gateway: Optional[str] = None
    ) -> None:
        self.name: Optional[str] = technical_name
        self.adapter_type: Optional[str] = adapter_type
        self.dns_suffix: Optional[str] = dns_suffix
        self.ipv4_address: Optional[str] = ipv4_address
        self.ipv6_address: Optional[str] = ipv6_address
        self.subnet_mask: Optional[str] = subnet_mask
        self.default_gateway: Optional[str] = default_gateway

    def __str__(self):
        indent = 2 * ' '
        formatted_info = [
            self.get_heading(),
            '',
            f'{indent}{self.get_interface_type()}' if self.name is not None else '',
            f'{indent}{self.get_dns_suffix()}' if self.dns_suffix is not None else '',
            f'{indent}{self.get_ipv4_addr()}' if self.ipv4_address is not None else '',
            f'{indent}{self.get_ipv6_addr()}' if self.ipv6_address is not None else '',
            f'{indent}{self.get_subnet_mask()}' if self.subnet_mask is not None else '',
            f'{indent}{self.get_formatted_default_gateway()}' if self.default_gateway is not None else ''
        ]


        non_empty_info = filter(lambda x: x != '', formatted_info)

        return "\n".join(non_empty_info)

    def get_heading(self):
        friendly_name = self.get_friendly_interface_name(self.name)
        return friendly_name

    def get_interface_type(self):
        return f"{self.get_formatted_label('Type')}{self.adapter_type}"

    def get_dns_suffix(self):
        return f"{self.get_formatted_label('DNS Suffix')}{self.dns_suffix}"

    def get_ipv4_addr(self):
        return f"{self.get_formatted_label('IPv4 Address')}{self.ipv4_address}"

    def get_ipv6_addr(self):
        return f"{self.get_formatted_label('IPv6 Address')}{self.ipv6_address}"

    def get_subnet_mask(self):
        try:
            the_int = int(self.subnet_mask)
            self.subnet_mask = self.cidr_to_netmask(the_int)
        except:
            return ''
        return f"{self.get_formatted_label('Subnet Mask')}{self.subnet_mask}"

    def get_formatted_default_gateway(self):
        return f"{self.get_formatted_label('Default Gateway')}{self.default_gateway}"


    @staticmethod
    def get_formatted_label(label):
        target_size = 30
        blank_padding = " " * (target_size - len(label))
        dotted_padding = ''.join([c if (i+len(label)) % 2 == 0 else '.' for i, c in enumerate(blank_padding)])
        adjusted_label = f"{label}{dotted_padding}"[:target_size] + " : "
        return adjusted_label

    @staticmethod
    def get_friendly_interface_name(interface_name : str) -> str:
        if interface_name.startswith('en'):
            return "Ethernet Adapter"
        elif interface_name.startswith('wl'):
            return "Wireless LAN Adapter"
        elif interface_name.startswith('virbr'):
            return "Virtual Bridge Interface"
        elif interface_name.startswith('veth'):
            return "Virtual Ethernet Interface"
        elif interface_name.startswith('tun') or interface_name.startswith('tap'):
            return "VPN Tunnel Interface"
        elif interface_name.startswith('bond'):
            return "Bonding Interface"
        elif '.' in interface_name:
            return "VLAN Interface"
        elif interface_name == 'lo':
            return "Loopback Interface"
        elif interface_name.startswith('br'):
            return "Bridge Interface"
        else:
            return f"Unknown Adapter ({interface_name})"

    @staticmethod
    def cidr_to_netmask(cidr_bits: int) -> str:
        mask = ('1' * cidr_bits).ljust(32, '0')
        octets = [str(int(mask[i:i + 8], 2)) for i in range(0, 32, 8)]
        return '.'.join(octets)



from __future__ import annotations
from dataclasses import dataclass
from typing import Union

# ----------------------------------------------

@dataclass
class Grid:
    x_size : int
    y_size : int

    def get_num(self, pt : LatticePoint) -> int:
        if not self.is_in_bounds(pt):
            raise ValueError(f"Point {pt} is outside of the grid bounds")

        return pt.y * self.x_size + pt.x

    def get_pt(self, num : int) -> LatticePoint:
        if num < 0 or num >= self.x_size * self.y_size:
            raise ValueError(f"Number {num} is outside of the grid bounds")
        return LatticePoint(num % self.x_size, num // self.x_size)

    def get_lattice_vectors(self) -> list[Vector]:
        return [lattice_point.to_vector() for lattice_point in self.get_lattice_points()]

    def get_lattice_points(self) -> list[LatticePoint]:
        return [LatticePoint(x, y) for x in range(self.x_size) for y in range(self.y_size)]

    def in_horizontal_bounds(self, x : int) -> bool:
        return 0 <= x <= self.x_size

    def is_in_vertical_bounds(self, y : int) -> bool:
        return 0 <= y <= self.y_size

    def is_in_bounds(self, vec : Union[LatticePoint, Vector]) -> bool:
        return self.in_horizontal_bounds(vec.x) and self.is_in_vertical_bounds(vec.y)


@dataclass
class Vector:
    x : float
    y : float

    def as_tuple(self):
        return self.x, self.y

    def __add__(self, other : Vector):
        return Vector(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Vector(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        return Vector(self.x * other, self.y * other)

    def to_lattice(self) -> LatticePoint:
        return LatticePoint(round(self.x), round(self.y))


@dataclass
class LatticePoint:
    x: int
    y: int

    def to_vector(self) -> Vector:
        return Vector(x=self.x, y=self.y)

    def as_tuple(self) -> (int, int):
        return self.x, self.y

    def __mul__(self, other : float):
        return LatticePoint(int(self.x * other), int(self.y * other))

    def __add__(self, other : LatticePoint):
        if isinstance(other, LatticePoint):
            return LatticePoint(self.x + other.x, self.y + other.y)
        else:
            raise TypeError("Can only add LatticePoint to another LatticePoint")

    def __sub__(self, other : LatticePoint):
        if isinstance(other, LatticePoint):
            return LatticePoint(self.x - other.x, self.y - other.y)
        else:
            raise TypeError("Can only subtract LatticePoint from another LatticePoint")

@dataclass
class Click:
    point : LatticePoint
    display_index : int

    def to_str(self):
        return f'{self.point.x},{self.point.y},{self.display_index}'

    @classmethod
    def from_str(cls, s : str) -> Click:
        x, y, display_index = s.split(',')
        return cls(point=LatticePoint(x=int(x), y=int(y)), display_index=int(display_index))


if __name__ == "__main__":
    point1 = LatticePoint(3, 5)
    point2 = LatticePoint(1, 2)
    point3 = point1 + point2  # This will be LatticePoint(x=4, y=7)
    point4 = point1 - point2  # This will be LatticePoint(x=2, y=3)

    print(point4, point3)



from __future__ import annotations

import os
from abc import abstractmethod, ABC
from typing import Optional

from holytools.logging import Timber, LogLevel


# ---------------------------------------------------------

class BaseConfigs(Timber, ABC):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._map = {None : {}}
        self._populate_map()

    @abstractmethod
    def _populate_map(self):
        pass

    @staticmethod
    def _as_abspath(path: str) -> str:
        path = os.path.expanduser(path=path)
        path = os.path.abspath(path)
        return path

    # ---------------------------------------------------------
    # interface

    def get(self, key : str, section : Optional[str] = None) -> Optional[str]:
        try:
            config_value = self._map[section][key]
        except KeyError:
            self.log(msg=f'Could not find key \"{key}\" under section \"{section}\" in configs', level=LogLevel.WARNING)
            config_value = None

        return config_value

    def set(self, key : str, value : str, section : Optional[str] = None):
        if not section in self._map:
            self._map[section] = {}
        if key in self._map[section]:
            raise ValueError(f'Key \"{key}\" already exists in settings')

        self._map[section][key] = value
        self._update_resource()

    def remove(self, key : str, section : Optional[str] = None):
        if key in self._map[section]:
            del self._map[section][key]
            self._update_resource()

    def get_general_section(self) -> dict[str, str]:
        return self._map[None]

    @abstractmethod
    def _update_resource(self):
        pass






from .tools import LoggingTools, CaptureLogs
from .loggers import LoggerOverseer, LoggerFactory
from .timber import Timber


import logging
class LogLevel:
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


import os
import sys
import traceback
import unittest
from unittest import case


class DirectoryLimitedLoader(unittest.TestLoader):
    def loadTestsFromModule(self, module, *, pattern=None):
        tests = []
        for name in dir(module):
            obj = getattr(module, name)
            if isinstance(obj, type) and issubclass(obj, case.TestCase) and obj not in (case.TestCase, case.FunctionTestCase):
                tests_dirpath = os.path.join(ROOT_DIRPATH, 'tests')

                if not tests_dirpath in sys.modules[obj.__module__].__file__:
                    continue

                tests.append(self.loadTestsFromTestCase(obj))

        load_tests = getattr(module, 'load_tests', None)
        # noinspection PyTypeChecker
        tests = self.suiteClass(tests)
        if load_tests is not None:
            try:
                return load_tests(self, tests, pattern)
            except Exception as e:
                error_case, error_message = _make_failed_load_tests(
                    module.__name__, e, self.suiteClass)
                self.errors.append(error_message)
                return error_case
        return tests

def _make_failed_load_tests(name, exception, suiteClass):
    message = 'Failed to call load_tests:\n%s' % (traceback.format_exc(),)
    return _make_failed_test(
        name, exception, suiteClass, message)

def _make_failed_test(methodname, exception, suiteClass, message):
    test = _FailedTest(methodname, exception)
    return suiteClass((test,)), message

class _FailedTest(case.TestCase):
    _testMethodName = None

    def __init__(self, method_name, exception):
        self._exception = exception
        super(_FailedTest, self).__init__(method_name)

    def __getattr__(self, name):
        if name != self._testMethodName:
            # noinspection PyUnresolvedReferences
            return super(_FailedTest, self).__getattr__(name)
        def testFailure():
            raise self._exception
        return testFailure


def run_unittests(start_dir : str, pattern : str):
    loader = DirectoryLimitedLoader()
    suite = loader.discover(start_dir=start_dir, pattern=pattern)
    runner = unittest.TextTestRunner()
    runner.run(suite)


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument('-s', required=True)
    parser.add_argument('-p', required=True)
    parser.add_argument('-root', required=True)

    args = parser.parse_args()
    ROOT_DIRPATH = args.root
    run_unittests(start_dir=args.s, pattern=args.p)



import time
from queue import Queue

from pynput.keyboard import Key as PynputKey
from pynput.keyboard import KeyCode
from pynput.keyboard import Controller as KeyboardController
from pynput import keyboard
from typing import Union, Any, Optional
import atexit

Key = Union[PynputKey, KeyCode]

# ----------------------------------------------


class InputWaiter:
    def __init__(self, target_value : Optional[Any] = None):
        self.q = Queue()
        self.target_value : Optional[Any] = target_value
        self.is_done : bool = False

    def clear(self):
        self.q = Queue()

    def write(self, value : Optional[Any] = None):
        self.q.put(value)

    def get(self) -> Any:
        while True:
            value = self.q.get()
            if self.target_value == value:
                self.is_done = True
                return value


class Keyboard:
    def __init__(self):
        self._keyboard = KeyboardController()

        def del_keyboard():
            del self._keyboard
        atexit.register(del_keyboard)

    def type(self, msg: str):
        for char in msg:
            self._keyboard.type(char)  # Type the character
            time.sleep(0.02)  # Interval between keys



class KeyboardListener:
    def __init__(self, verbose : bool = False):
        self.verbose = verbose

        self.listener = keyboard.Listener(on_press=self._on_press, on_release=self._on_release)
        self.listener.start()
        self.pressed_buttons : set[Key] = set()
        self.press_waiters : list[InputWaiter] = []
        self.release_waiters : list[InputWaiter] = []

    def wait_on_hold(self, key : Key, duration : float):
        while True:
            press_waiter = self._register_press_waiter(target_value=key)
            press_waiter.get()
            if self.verbose:
                print(f'Press of key \"{key}\" registered. Hold for {duration} to finsh')

            time.sleep(duration)
            still_held  = self.check_key_pressed(key=key)
            if still_held:
                break

    def check_key_pressed(self, key : Key):
        return key in self.pressed_buttons

    def wait_on_press(self, key : Key):
        waiter = self._register_press_waiter(target_value=key)
        waiter.get()

    def get_next_key(self) -> Key:
        waiter = InputWaiter()
        self.press_waiters.append(waiter)
        return waiter.get()

    # ---------------------------------------------------------

    def _on_press(self, key: Key):
        if self.verbose:
            print(f'key press registered {key}')
        for waiter in self.press_waiters:
            waiter.write(key)
        self._remove_finished_waiters()
        self.pressed_buttons.add(key)


    def _on_release(self, key : Key):
        if self.verbose:
            print(f'key release registered {key}')
        if key in self.pressed_buttons:
            self.pressed_buttons.remove(key)
        for waiter in self.release_waiters:
            waiter.write(key)
        self._remove_finished_waiters()


    def _remove_finished_waiters(self):
        for waiter in self.press_waiters:
            if waiter.is_done:
                self.press_waiters.remove(waiter)
        for waiter in self.release_waiters:
            if waiter.is_done:
                self.release_waiters.remove(waiter)

    def _register_press_waiter(self, target_value : Key) -> InputWaiter:
        waiter = InputWaiter(target_value)
        self.press_waiters.append(waiter)
        return waiter


if __name__ == "__main__":
    # listener = KeyboardListener(verbose=True)
    test_keyboard = Keyboard()
    listener = KeyboardListener(verbose=True)
    listener.wait_on_hold(key=KeyCode.from_char(char='a'), duration=20)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from PIL import ImageDraw, Image, ImageFont
from PIL.Image import Image as PILImage
from .types import Grid, LatticePoint, Vector


# ----------------------------------------------

class Orientation(Enum):
    HORIZONTAL = 0
    VERTICAL = 1


class EditableImage:
    def __init__(self, img : PILImage, mapper : PixelMapper):
        self.pil_image : PILImage = img.convert('RGBA')

        ImageDraw.Draw(self.pil_image)
        self.mapper : PixelMapper = mapper
        self.overlay = Image.new('RGBA', self.pil_image.size, (255, 255, 255, 0))
        self.overlay = self.overlay.convert('RGBA')
        self.overlay_draw = ImageDraw.Draw(self.overlay)


    def get_grid_overlay(self) -> PILImage:
        self.draw_cell_labels()
        self.draw_grid_lines()
        return Image.alpha_composite(self.pil_image, self.overlay)


    def draw_cell_labels(self):
        grid = self.mapper.input_grid
        for pt in grid.get_lattice_points():
            input_vector = pt.to_vector() + Vector(x=0.5, y=0.5)
            output_vector = self.mapper.map_vector(vector=input_vector)
            output_pt = output_vector.to_lattice()
            text = str(grid.get_num(pt=pt))
            self.draw_text(text=text, point=output_pt, font_size=int(400 / grid.x_size))


    def draw_grid_lines(self):
        for x in range(0, self.mapper.input_grid.x_size + 1):
            x_px = self.mapper.map_horizontal(x=x)
            self.draw_line(coordinate=x_px, orientation=Orientation.VERTICAL)
        for y in range(0, self.mapper.input_grid.y_size + 1):
            y_px = self.mapper.map_vertical(y=y)
            self.draw_line(coordinate=y_px, orientation=Orientation.HORIZONTAL)


    def draw_line(self, coordinate : int, orientation : Orientation, opacity=128):
        color_vector = (255,0,0, opacity)
        if orientation == Orientation.HORIZONTAL:
            start_pos = (0, coordinate)
            end_pos = (self.overlay.width, coordinate)
        else:
            start_pos = (coordinate, 0)
            end_pos = (coordinate, self.overlay.height)

        self.overlay_draw.line([start_pos, end_pos], fill=color_vector, width=1)


    def draw_text(self, text: str, point: LatticePoint, font_size=20, opacity=255):
        font = ImageFont.load_default(size=font_size)
        delta = LatticePoint(font_size // 2, font_size // 2)
        centered_point = point - delta

        background_color = self.pil_image.getpixel(centered_point.as_tuple())
        brightness = sum(background_color[:3]) / (3 * 255)
        if brightness < 0.5:
            color = (255, 255,255 , opacity)
        else:
            color = (0, 0, 0, opacity)
        self.overlay_draw.text(centered_point.as_tuple(), text=text, fill=color, font=font)


@dataclass
class PixelMapper:
    input_grid : Grid
    output_grid : Grid

    def map_vector(self, vector : Vector):
        if not self.input_grid.is_in_bounds(vec=vector):
            raise ValueError(f"Vector {vector} is outside of the grid bounds")

        return Vector(x=self.map_horizontal(vector.x), y=self.map_vertical(vector.y))

    def map_pt(self, point : LatticePoint) -> LatticePoint:
        if not self.input_grid.is_in_bounds(vec=point):
            raise ValueError(f"Lattice point {point} is outside of the grid bounds")
        return LatticePoint(x=self.map_horizontal(point.x), y=self.map_vertical(point.y))

    def map_horizontal(self, x : float) -> int:
        if not self.input_grid.in_horizontal_bounds(x=round(x)):
            raise ValueError(f"X value {x} is outside of the grid bounds")
        return round(x * self.output_grid.x_size / self.input_grid.x_size)

    def map_vertical(self, y : float) -> int:
        if not self.input_grid.is_in_vertical_bounds(y=round(y)):
            raise ValueError(f"Y value {y} is outside of the grid bounds")
        return round(y * self.output_grid.y_size / self.input_grid.y_size)



from __future__ import annotations

import os


# -------------------------------------------

class ResourceManager:
    def __init__(self, root_dirpath : str):
        if not os.path.isdir(root_dirpath):
            raise ValueError(f'Failed to initialize {self.__class__.__name__}: Root directory {root_dirpath} does not exist')
        self._root_dirpath : str = root_dirpath
        self._subdir_paths: list[str] = []
        self._fpaths : list[str] = []

    def add_dir(self, relative_path: str) -> str:
        dirpath = self._get_relative_path(relative_path=relative_path)
        if not os.path.isdir(dirpath):
            os.makedirs(dirpath, exist_ok=True)
        self._subdir_paths.append(dirpath)
        return dirpath

    def add_file(self, relative_path: str) -> str:
        fpath = self._get_relative_path(relative_path=relative_path)
        self._fpaths.append(fpath)
        return fpath

    # -------------------------------------------

    def _get_relative_path(self, relative_path: str) -> str:
        root_dirpath = self.get_root_dirpath()
        return os.path.join(root_dirpath, relative_path)

    def get_root_dirpath(self) -> str:
        if self._root_dirpath is None:
            raise ValueError(f'Root directory not set for {self.__class__.__name__}')
        return self._root_dirpath



from __future__ import annotations

import ctypes
import inspect
import multiprocessing
import threading
import time
import unittest
import unittest.mock
import unittest.mock
from abc import abstractmethod
from multiprocessing import Process, Value
from typing import Optional, Callable

from holytools.devtools.testing.result import Result, Report, CaseStatus
from .basetest import BaseTest
from .runner import Runner


# ---------------------------------------------------------

class Unittest(BaseTest):
    @classmethod
    def ready(cls) -> Unittest:
        instance = cls()
        instance.setUpClass()
        instance.setUp()
        return instance

    @classmethod
    def execute_all(cls, manual_mode : bool = True, trace_resourcewarning : bool = False):
        suite = unittest.TestLoader().loadTestsFromTestCase(cls)
        runner = Runner(logger=cls.get_logger(), is_manual=manual_mode, test_name=cls.__name__)
        tracemalloc_depth = 10 if trace_resourcewarning else 0
        results = runner.run(testsuite=suite, tracemalloc_depth=tracemalloc_depth)

        return results

    @classmethod
    def execute_stats(cls, reps : int, min_success_percent : float, test_names : Optional[list[str]] = None):
        if not 0 <= min_success_percent <= 100:
            raise ValueError(f'Min success percent should be in [0,100], got {min_success_percent}')

        test_names = test_names or unittest.TestLoader().getTestCaseNames(cls)
        stat_reports : list[Report] = []
        outcome_map : dict[str, list] = {}

        for tn in test_names:
            report_name = f'{cls.__name__}.{tn}'
            start_time = time.time()
            suite_result = cls._run_several(reps=reps, name=tn)
            runtime = round(time.time() - start_time, 3)

            outcomes = [True if c.status == CaseStatus.SUCCESS else False for c in suite_result.reports]
            outcome_map[report_name] = outcomes
            success_pc = 100*sum(outcomes) / len(outcomes)

            status = CaseStatus.SUCCESS if success_pc >= min_success_percent else CaseStatus.FAIL
            if suite_result.reports[0].status == CaseStatus.ERROR:
                status = CaseStatus.ERROR
            stats_report = Report(name=report_name, status=status, runtime=runtime)
            stat_reports.append(stats_report)

        result = Result(logger=cls.get_logger(), testsuite_name=cls.__name__)
        spaces = 13
        for c in stat_reports:
            outcomes = outcome_map[c.name]
            num_successful, total = sum(outcomes), len(outcomes)
            
            result.log_test_start(casename=c.name)
            status_arr = ['✓' if o else '✗' for o in outcomes]
            result.log(f'{"Result":<{spaces}}: {status_arr}')
            result.log(f'{"Success rate":<{spaces}}: {num_successful/total*100}%')
            result.log(msg=f'{"Status":<{spaces}}: {c.status}\n', level=c.get_log_level())

        result.reports = stat_reports
        result.log_summary()

    @classmethod
    def _run_several(cls, name : str, reps : int) -> Result:
        suite = unittest.TestSuite()
        current_case = cls(name)
        for _ in range(reps):
            suite.addTest(current_case)

        runner = Runner(logger=cls.get_logger(), test_name=cls.__name__)
        results = runner.run(testsuite=suite)

        return results

    @staticmethod
    def patch_module(original: type | Callable, replacement: type | Callable):
        module_path = inspect.getmodule(original).__name__
        qualified_name = original.__qualname__
        frame = inspect.currentframe().f_back
        caller_module = inspect.getmodule(frame)

        try:
            # corresponds to "from [caller_module] import [original]
            _ = getattr(caller_module, qualified_name)
            full_path = f"{caller_module.__name__}.{qualified_name}"
        except Exception:
            # corresponds to import [caller_module].[original]
            full_path = f"{module_path}.{qualified_name}"

        # print(f'Full path = {full_path}')
        def decorator(func):
            return unittest.mock.patch(full_path, replacement)(func)

        return decorator


class BlockedTester:
    def __init__(self):
        self.shared_bool = Value(ctypes.c_bool, False)

    def check_ok(self, case : str, delay : float) -> bool:
        def do_run():
            threading.Thread(target=self.blocked).start()
            time.sleep(delay)
            check_thread = threading.Thread(target=self.check_condition, args=(case,))
            check_thread.start()
            check_thread.join()
            q.put('stop')

        q = multiprocessing.Queue()
        process = Process(target=do_run)
        process.start()
        q.get()
        process.terminate()
        return self.shared_bool.value


    @abstractmethod
    def blocked(self):
        pass

    def check_condition(self, case : str):
        self.shared_bool.value = self.perform_check(case=case)

    @abstractmethod
    def perform_check(self, case : str) -> bool:
        pass




from __future__ import annotations

from typing import Optional
from PIL import Image
from mss import mss
from screeninfo import get_monitors
from screeninfo import Monitor as BaseMonitor

from .grid_draw import EditableImage, PixelMapper
from .types import Grid, LatticePoint


# ----------------------------------------------


class Display(BaseMonitor):
    @classmethod
    def get_primary(cls) -> Optional[Display]:
        for monitor in get_monitors():
            if monitor.is_primary:
                return cls.from_base(base_monitor=monitor)
        return None

    @classmethod
    def get_secondary(cls) -> Optional[Display]:
        for monitor in get_monitors():
            if not monitor.is_primary:
                return cls.from_base(base_monitor=monitor)
        return None

    @classmethod
    def from_base(cls, base_monitor : BaseMonitor) -> Display:
        return cls(x=base_monitor.x, y=base_monitor.y, width=base_monitor.width, height=base_monitor.height, is_primary=base_monitor.is_primary)

    # ----------------------------------------------

    def get_screenshot(self, grid : Optional[Grid] = None):
        with mss() as sct:
            monitor_dict = {"top": self.y, "left": self.x, "width": self.width, "height": self.height}
            sct_img = sct.grab(monitor_dict)
            img = Image.frombytes('RGB', sct_img.size, sct_img.bgra, 'raw', 'BGRX')
        if grid:
            editable = EditableImage(img=img, mapper=self.get_mapper(grid=grid))
            img = editable.get_grid_overlay()

        return img

    def get_mapper(self, grid : Grid):
        pixel_grid = Grid(x_size=self.width, y_size=self.height)
        return PixelMapper(input_grid=grid, output_grid=pixel_grid)

    def to_virtual_display(self, pixel : LatticePoint) -> LatticePoint:
        origin = LatticePoint(x=self.x, y = self.y)
        return origin + pixel

    def in_bounds(self, pixel : LatticePoint):
        return 0 <= pixel.x <= self.width and 0 <= pixel.y <= self.height

    def is_horizontal(self):
        return self.width > self.height

    def is_vertical(self):
        return self.height > self.width



from __future__ import annotations
from typing import Optional
from abc import abstractmethod, ABC

# -------------------------------------------


class TreeNode(ABC):
    def get_subnodes(self, *args, **kwargs) -> list[TreeNode]:
        _, __ = args, kwargs
        subnodes = []
        for child in self.get_child_nodes():
            subnodes.append(child)
            subnodes += child.get_subnodes()
        return subnodes

    def get_ancestors(self) -> list[TreeNode]:
        current = self
        ancestors = []
        while current.get_parent():
            ancestors.append(current.get_parent())
            current = current.get_parent()
        return ancestors

    def get_root(self) -> TreeNode:
        current = self
        while current.get_parent():
            current = current.get_parent()
        return current

    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    def get_child_nodes(self, *args, **kwargs) -> list[TreeNode]:
        pass

    @abstractmethod
    def get_parent(self) -> Optional[TreeNode]:
        pass

    def __str__(self):
        return self.get_name()


from datetime import datetime
from threading import Event
from typing import Callable
from .scheduler import TaskScheduler

class Countdown:
    def __init__(self, duration: float, on_expiration: Callable = lambda *args, **kwargs: None):
        self.duration : float = duration
        self.on_expiration : Callable = on_expiration
        self.scheduler : TaskScheduler = TaskScheduler()
        self.one_time_lock = Lock()

    def start(self):
        self.scheduler.submit_once(task=self._release, delay=self.duration)

    def restart(self):
        self.scheduler.cancel_all()
        self.start()

    def is_active(self):
        return self.scheduler.is_active()

    def wait(self):
        self.one_time_lock.wait()

    # -------------------------------------------

    def _release(self):
        self.one_time_lock.unlock()
        self.on_expiration()


class Lock:
    def __init__(self):
        self._event = Event()
        self._event.clear()

    def wait(self):
        self._event.wait()

    def unlock(self):
        self._event.set()



class Timer:
    def __init__(self):
        self.start_time : datetime = datetime.now()

    def restart(self):
        self.start_time = datetime.now()

    def capture(self, verbose : bool = True) -> float:
        now = datetime.now()
        delta = now-self.start_time
        delta_sec = delta.total_seconds()
        if verbose:
            print(f'Time has been running for {delta_sec} seconds')
        return delta_sec



import time

from pynput.mouse import Button
from pynput.mouse import Controller as MouseController
from holytools.hardware.display import Display, LatticePoint, Vector
from holytools.hardware.display import Grid
from typing import Optional

# ----------------------------------------------

class Mouse:
    def __init__(self):
        self._mouse : MouseController = MouseController()

    def click(self, pixel_x : int, pixel_y : int, on_primary_display: bool = True):
        point = LatticePoint(pixel_x, pixel_y)
        display = Display.get_primary() if on_primary_display else Display.get_secondary()
        if not display.in_bounds(point):
            raise ValueError(f"Point {point} is outside of the display bounds")
        rel_to_primary = display.to_virtual_display(pixel=point)

        self._mouse.position = (rel_to_primary.x, rel_to_primary.y)
        self._mouse.click(Button.left, 1)

class TextMouse:
    def __init__(self, input_grid: Grid = Grid(x_size=25, y_size=25)):
        self.mouse : Mouse = Mouse()
        self.input_grid : Grid = input_grid
        self.primary : Display = Display.get_primary()
        self.secondary : Optional[Display] = Display.get_secondary()

    def get_view(self, on_primary_display : bool = True):
        display = self.primary if on_primary_display else self.secondary
        return display.get_screenshot(grid=self.input_grid)

    def click(self, cell_num : int, on_primary_display : bool = True):
        left_corner = self.input_grid.get_pt(num=cell_num).to_vector()
        center = left_corner + Vector(0.5,0.5)

        display = self.primary if on_primary_display else self.secondary
        if not display:
            raise ValueError("There is no secondary display")

        mapper = display.get_mapper(grid=self.input_grid)
        center = mapper.map_vector(center).to_lattice()
        self.mouse.click(center.x, center.y, on_primary_display=on_primary_display)


if __name__ == "__main__":
    text_mouse = TextMouse()
    while True:
        view = text_mouse.get_view()
        view.show()
        num = int(input(f'Click on cell:'))
        text_mouse.click(num)
        time.sleep(1)
        
        



# import speech_recognition as sr
# from queue import Queue
# import threading
#
#
# class BytePipe(Queue):
#     def get(self, block = True, timeout = None) -> bytes:
#         return super().get(block, timeout)
#
#     def put(self, item : bytes, block = True, timeout = None):
#         super().put(item, block, timeout)
#
#
# class Recorder:
#     def __init__(self):
#         self.recognizer = sr.Recognizer()
#         self.recognizer.non_speaking_duration = 0.1
#         self.recognizer.pause_threshold = 0.2
#         self.recognizer.energy_threshold = 750
#
#         self.is_running : bool = False
#         self.pipes: list[BytePipe] = []
#
#     def register_pipe(self) -> BytePipe:
#         pipe = BytePipe()
#         self.pipes.append(pipe)
#         return pipe
#
#     def start(self):
#         def do():
#             self.is_running = True
#             with sr.Microphone() as source:
#                 self.listen(source=source)
#         threading.Thread(target=do).start()
#
#     def stop(self):
#         self.is_running = False
#
#     def listen(self, source : sr.AudioSource):
#         while True:
#             if not self.is_running:
#                 break
#             audio_data = self.recognizer.listen(source)
#             wav_bytes = audio_data.get_wav_data()
#             for pipe in self.pipes:
#                 pipe.put(wav_bytes)
#
#
#



from abc import abstractmethod
from typing import Optional
import os
from enum import Enum
from pathlib import Path
from base64 import b64encode
from abc import ABC

class Access(Enum):
    READABLE = os.R_OK
    WRITABLE = os.W_OK
    EXECUTABLE = os.X_OK


class FileIO(ABC):
    def __init__(self, fpath : str, require_writable : bool = True, require_executable : bool = False):
        self.fpath : str = str(Path(fpath).absolute())
        required_permissions : set[Access] = {Access.READABLE}
        if require_writable:
            required_permissions.add(Access.WRITABLE)
        if require_executable:
            required_permissions.add(Access.EXECUTABLE)

        missing_permissions = [access for access in required_permissions if not self.has_permission(access=access)]
        if missing_permissions:
            raise PermissionError(f'File {fpath} does not have required permissions: {missing_permissions}')

        if self.exists_on_disk():
            self.check_content_ok()

    def has_permission(self, access : Access) -> bool:
        if os.path.isdir(self.fpath):
            return False
        if os.path.isfile(self.fpath):
            return os.access(self.fpath, access.value)
        else:
            parent_directory = os.path.dirname(self.fpath)
            return os.access(parent_directory, access.value) if parent_directory else False

    @abstractmethod
    def check_content_ok(self):
        pass

    @abstractmethod
    def read(self):
        pass

    @abstractmethod
    def write(self, content):
        pass

    @abstractmethod
    def view(self):
        pass

    def as_bytes(self):
        with open(self.fpath, 'rb') as file:
            content = file.read()
        return content

    def as_base64(self) -> str:
        data = self.as_bytes()
        return b64encode(data).decode(encoding='utf-8')

    def exists_on_disk(self) -> bool:
        return os.path.isfile(self.fpath)

    # ------------------------------------------------------

    def _get_suffix(self) -> Optional[str]:
        parts = self.fpath.split('.')
        suffix = parts[-1] if len(parts) > 1 else None
        return suffix


import os
import shutil
import tempfile

from holytools.fileIO.filetypes import PlaintextFile, BinaryFile, ImageFile

# ---------------------------------------------------------

def create_temp_copy(filename: str) -> str:
    module_dir = os.path.dirname(__file__)
    original_file_path = os.path.join(module_dir, filename)

    temp_fd, temp_filepath = tempfile.mkstemp(suffix=os.path.splitext(filename)[1], dir='/tmp')
    os.close(temp_fd)
    shutil.copy2(original_file_path, temp_filepath)

    return temp_filepath


class ExampleFiles:
    @staticmethod
    def lend_png() -> ImageFile:
        fpath = create_temp_copy('file.png')
        return ImageFile(fpath=fpath)

    @staticmethod
    def lend_jpg() -> ImageFile:
        fpath = create_temp_copy('file.jpg')
        return ImageFile(fpath=fpath)

    @staticmethod
    def lend_pdf() -> PlaintextFile:
        fpath = create_temp_copy('file.pdf')
        return PlaintextFile(fpath=fpath)

    @staticmethod
    def lend_txt() -> PlaintextFile:
        fpath = create_temp_copy('file.txt')
        return PlaintextFile(fpath=fpath)

    @staticmethod
    def lend_csv() -> PlaintextFile:
        fpath = create_temp_copy('file.csv')
        return PlaintextFile(fpath=fpath)

    @staticmethod
    def lend_bin() -> BinaryFile:
        fpath = create_temp_copy('file.bin')
        return BinaryFile(fpath=fpath)

    @staticmethod
    def lend_wav() -> BinaryFile:
        fpath = create_temp_copy('file.wav')
        return BinaryFile(fpath=fpath)


if __name__ == "__main__":
    print(os.path.abspath(__file__))


from __future__ import annotations

import psutil
# -------------------------------------------


class RAM:
    @staticmethod
    def get_available_in_GB(include_swap : bool = False) -> float:
        available_memory = psutil.virtual_memory().available / (1024 ** 3)
        if include_swap:
            swap_memory = psutil.swap_memory().free / (1024 ** 3)
            available_memory += swap_memory
        return available_memory

    @staticmethod
    def get_total_in_GB(include_swap : bool = False) -> float:
        total_memory = psutil.virtual_memory().total / (1024 ** 3)
        if include_swap:
            swap_memory = psutil.swap_memory().total / (1024 ** 3)
            total_memory += swap_memory

        return total_memory


if __name__ == '__main__':
    total = RAM().get_total_in_GB(include_swap=True)


import datetime
import subprocess  # For executing a shell command
import time
import tabulate


# ----------------------------

class ConnectivityTester:
    def __init__(self, remote_ip_addr : str = '172.217.164.99', remote_name : str = 'google.de', package_size : int = 256):
        self.remote_ip_addr : str = remote_ip_addr
        self.remote_name : str = remote_name
        self.package_size_bytes : int = package_size

    def check_connectivity(self, max_duration : int = 30, verbose : bool = False):
        hosts = [self.remote_ip_addr, self.remote_name]
        ping_timeout = 1
        poll_duration = ping_timeout*len(hosts)

        ping_results = {}
        latencies = {}
        for h in hosts:
            ping_results[h] = []
            latencies[h] = []

        print(f'Launching connectivity tests for {self.remote_name} = {self.remote_ip_addr}. '
              f'\n  - Duration = {max_duration}s'
              f'\n  - Timeout = {ping_timeout}s'
              f'\n  - Package size = {self.package_size_bytes} bytes\n')
        start_time = time.time()
        while time.time() - start_time < max_duration:
            for h in hosts:
                is_reachable = self.is_ping_reachable(ping_timeout, host=h)
                remaining_time = ping_timeout - (time.time() - start_time) % ping_timeout

                ping_results[h].append(is_reachable)
                latencies[h].append(ping_timeout-remaining_time)
                if remaining_time > 0:
                    time.sleep(remaining_time)

            if not verbose:
                continue

            timestamp = datetime.datetime.now()
            timestamp = timestamp.replace(microsecond=0)
            result_msg = f'Success' if all(ping_results[h][-1] for h in hosts) else 'Failure'
            print(f'\nConnectivity results on {timestamp}: {result_msg}')
            for h in hosts:
                ip_reachibility, latency = ping_results[h][-1], latencies[h][-1]
                msg = f'✓ | Latency: {round(latency*1000)} ms' if ip_reachibility else f'✗ (Timeout after {ping_timeout*1000}ms)'
                print(f'    - Connection to {h:<16} {msg}')

        print(f'Summary of connectivity results:')
        table_data = []
        for h in hosts:
            polled_duration = poll_duration*len(ping_results[h])

            uptime = sum(ping_results[h])*poll_duration
            average_latency = f'{sum(latencies[h])/len(latencies[h])*1000:.2f}ms'
            row = [h, f'{uptime}/{polled_duration}s', average_latency]
            table_data.append(row)
        col_headers = ['Host', 'Uptime', 'Average latency']
        print(tabulate.tabulate(tabular_data=table_data, headers=col_headers, tablefmt='psql'))
        print()

    def is_ping_reachable(self, timeout : int, host : str) -> bool:
        if timeout < 1:
            raise ValueError('Timeout must be at least 1 second')

        command = f'ping -w {timeout} -c 1 {host} -s {self.package_size_bytes}'
        return subprocess.call(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True) == 0


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--duration', type=int)
    args = parser.parse_args()


    if not args.duration:
        raise ValueError('Please provide a duration for the connectivity test')

    tester0 = ConnectivityTester(remote_name='chatgpt.com', package_size=512)


import linecache
import os
import tracemalloc
import unittest
import unittest.mock
import warnings
from logging import Logger
from unittest import TestSuite

from .result import Result


# ----------------------------------------------

class Runner(unittest.TextTestRunner):
    def __init__(self, logger : Logger, test_name : str, is_manual : bool = False, mute : bool = False):
        super().__init__(resultclass=None)
        self.logger : Logger = logger
        self.manual_mode : bool = is_manual
        self.test_name : str = test_name
        self.mute : bool = mute

    def run(self, testsuite : TestSuite, tracemalloc_depth : int = 0) -> Result:
        if tracemalloc_depth > 0:
            tracemalloc.start(tracemalloc_depth)


        with warnings.catch_warnings(record=True) as captured_warnings:
            warnings.simplefilter("ignore")
            warnings.simplefilter("always", ResourceWarning)

            result = Result(logger=self.logger,
                            testsuite_name=self.test_name,
                            stream=self.stream,
                            descriptions=self.descriptions,
                            verbosity=2,
                            manual_mode=self.manual_mode)
            result.startTestRun()
            testsuite(result)
            result.stopTestRun()
            result.printErrors()

        for warning in captured_warnings:
            if tracemalloc_depth > 0:
                print(f'- Unclosed resources:')
                print(Runner.warning_to_str(warning_msg=warning))
            else:
                self.logger.warning(msg=f'[Warning]: Unclosed resource: \"{warning.message}\."'
                                        f'Enable trace_resourcewarning to obtain object trace')

        warnings.simplefilter("ignore", ResourceWarning)
        if tracemalloc_depth > 0:
            tracemalloc.stop()

        return result

    @staticmethod
    def warning_to_str(warning_msg: warnings.WarningMessage) -> str:
        tb = tracemalloc.get_object_traceback(warning_msg.source)
        frames = list(tb)
        frames = [f for f in frames if Runner.is_relevant(frame=f)]

        result = ''
        for f in frames:
            file_path = f.filename
            line_number = f.lineno
            result += (f'File "{file_path}", line {line_number}\n'
                      f'    {linecache.getline(file_path, line_number).strip()}\n')
        return result

    @staticmethod
    def is_relevant(frame):
        not_unittest = not os.path.dirname(unittest.__file__) in frame.filename
        not_custom_unittest = not os.path.dirname(__file__) in frame.filename
        return not_unittest and not_custom_unittest





from .fileio import FileIO


class BinaryFile(FileIO):
    def read(self) -> bytes:
        return self.as_bytes()

    def write(self,content: bytes):
        with open(self.fpath, 'wb') as file:
            file.write(content)

    def view(self):
        content = self.read()
        hex_content = content.hex()
        for i in range(0, len(hex_content), 20):
            line = ' '.join(hex_content[j:j + 2] for j in range(i, min(i + 20, len(hex_content)), 2))
            print(line)

    def check_content_ok(self):
        pass






import time
import threading
import inspect
import uuid
from typing import Callable
from dataclasses import dataclass

@dataclass
class FuncTask:
    func : Callable
    is_canceled : bool = False

    def __post_init__(self):
        self.id: str = str(uuid.uuid4())

    def run(self):
        if not self.is_canceled:
            self.func()

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id


class TaskScheduler:
    def __init__(self):
        self.scheduled_tasks: dict[str, FuncTask] = {}
    
    def submit_once(self, task: Callable, delay: float) -> FuncTask:
        return self._schedule_task(task=task, delay=delay)

    def submit_periodic(self, task: Callable, interval: float):
        def periodic():
            threading.Thread(target=task).start()
            self.submit_periodic(task, interval)
        self._schedule_task(task=periodic, delay=interval)

    def cancel_all(self):
        for task in self.scheduled_tasks.values():
            task.is_canceled = True

    def is_active(self):
        return len(self.scheduled_tasks) != 0

    # ---------------------------------------------------------

    def _schedule_task(self, task : Callable, delay : float) -> FuncTask:
        parameters = inspect.signature(task).parameters.values()
        for param in parameters:
            not_args_or_kwargs = (param.kind not in [param.VAR_POSITIONAL, param.VAR_KEYWORD])
            has_no_defaults = (param.default is param.empty)
            if has_no_defaults and not_args_or_kwargs:
                raise InvalidCallableException("Cannot schedule task that requires arguments")

        task = FuncTask(func=task)
        def do_delayed():
            self.scheduled_tasks[task.id] = task
            time.sleep(delay)
            task.run()
            print(f'Scheduled tasks is {self.scheduled_tasks}')
            print(f'Task id is {task.id}')
            del self.scheduled_tasks[task.id]
        
        threading.Thread(target=do_delayed).start()
        return task


class InvalidCallableException(Exception):
    """Exception raised when a callable with arguments is passed where none are expected."""
    pass


if __name__ == "__main__":
    pass


from types import NoneType
from typing import get_origin, get_args, Union


class TypeInspector:
    @staticmethod
    def get_type_list(dtype : type) -> list[type]:
        """
        :return: If dtype is of form Union[<dtype>] returns [<dtype1>, <dtype2>,...] ; Else returns [<dtype>]
        """
        if get_origin(dtype) is Union:
            return list(get_args(dtype))
        else:
            return [dtype]

    # noinspection DuplicatedCode
    @staticmethod
    def get_core_type(dtype : type) -> type:
        """
        :return: If dtype is of form Optional[<dtype>] returns <dtype>; Else returns <dtype>
        """
        if get_origin(dtype) is Union:
            types = get_args(dtype)

            core_types = [t for t in types if not t is NoneType]
            if len(core_types) == 1:
                return core_types[0]
            else:
                raise ValueError(f'Union dtype {dtype} has more than one core dtype')
        else:
            return dtype

    @staticmethod
    def is_optional_type(dtype : type) -> bool:
        """
        :return: Returns true if <dtype> is of form Optional[<dtype>] or Union[None, (...)]; Else returns false
        """
        if get_origin(dtype) is Union:
            types = get_args(dtype)
            return any([t for t in types if t is NoneType])

        return False

    @staticmethod
    def check_dtype_confirmity(obj : object, dtype : type) -> bool:
        if get_origin(dtype) is list:
            if not isinstance(obj, list):
                return False
            if not get_args(dtype):
                return False
            element_type = get_args(dtype)[0]
            return all([isinstance(x, element_type) for x in obj])
        else:
            obj_conforms = isinstance(obj, dtype)
        return obj_conforms



import sys

from holytools.configs import FileConfigs


def main():
    configs = FileConfigs.credentials()
    num_args = len(sys.argv) - 1

    if num_args == 0:
        print(f'Credential store content')
        content = configs.read().strip()
        if len(content) == 0:
            return

        lines = content.split(f'\n')
        for l in lines:
            if len(l) > 0 and not l.startswith(f'['):
                l = f'├── {l}'
            print(l)

    elif num_args == 2:
        action, key = sys.argv[1], sys.argv[2]
        if action == 'insert':
            value = input(f'Enter value for key \"{key}\": ')
            configs.set(key=key, value=value)
        elif action == 'remove':
            configs.remove(key=key)
        else:
            raise ValueError(f'Unknown command \"{action}\"')
    else:
        raise ValueError(f'Invalid command syntax: {sys.argv}')

if __name__ == "__main__":
    main()


from __future__ import annotations

import logging
import unittest.mock
from typing import Optional

from holytools.logging import LoggerFactory


class BaseTest(unittest.TestCase):
    _logger: logging.Logger = None

    def __init__(self, *args):
        super().__init__(*args)
        self.is_manual_mode : bool = False

    def set_is_manual(self):
        self.is_manual_mode = True

    def get_is_manual(self):
        return self.is_manual_mode

    def get_name(self) -> str:
        full_test_name = self.id()
        parts = full_test_name.split('.')
        last_parts = parts[-2:]
        test_name = '.'.join(last_parts)
        return test_name

    @classmethod
    def get_logger(cls) -> logging.Logger:
        if not cls._logger:
            cls._logger = LoggerFactory.get_logger(include_location=False, include_timestamp=False, name=cls.__name__, use_stdout=True, log_fpath=cls.log_fpath())
        return cls._logger


    @classmethod
    def log_fpath(cls) -> Optional[str]:
        return None

    @classmethod
    def log(cls, msg : str, level : int = logging.INFO):
        cls.get_logger().log(msg=f'{msg}', level=level)

    @classmethod
    def warning(cls, msg : str, *args, **kwargs):
        kwargs['level'] = logging.WARNING
        cls._logger.log(msg=msg, *args, **kwargs)

    @classmethod
    def error(cls, msg : str, *args, **kwargs):
        kwargs['level'] = logging.ERROR
        cls._logger.log(msg=msg, *args, **kwargs)

    @classmethod
    def critical(cls, msg : str, *args, **kwargs):
        kwargs['level'] = logging.CRITICAL
        cls._logger.log(msg=msg, *args, **kwargs)

    @classmethod
    def info(cls, msg : str, *args, **kwargs):
        kwargs['level'] = logging.INFO
        cls._logger.log(msg=msg, *args, **kwargs)


    # ---------------------------------------------------------
    # assertions

    def assertEqual(self, first : object, second : object, msg : Optional[str] = None):
        if not first == second:
            first_str = str(first).__repr__()
            second_str = str(second).__repr__()
            if msg is None:
                msg = (f'Tested expressions should match:'
                       f'\nFirst : {first_str} ({type(first)})'
                       f'\nSecond: {second_str} ({type(second)})')
            raise AssertionError(msg)

    def assertSame(self, first : dict, second : dict, msg : Optional[str] = None):
        """Checks whether contents of dicts first and second are the same"""
        for key in first:
            first_obj = first[key]
            second_obj = second[key]
            self.assertSameElementary(type(first_obj), type(second_obj))
            if isinstance(first_obj, dict):
                self.assertSame(first_obj, second_obj, msg=msg)
            elif isinstance(first_obj, list):
                for i in range(len(first_obj)):
                    self.assertSameElementary(first_obj[i], second_obj[i])
            else:
                self.assertSameElementary(first_obj, second_obj)

    def assertSameElementary(self, first : object, second : object):
        if isinstance(first, float) and isinstance(second, float):
            self.assertSameFloat(first, second)
        else:
            self.assertEqual(first, second)

    @staticmethod
    def assertSameFloat(first : float, second : float, msg : Optional[str] = None):
        if first != first:
            same_float = second != second
        else:
            same_float = first == second
        if not same_float:
            first_str = str(first).__repr__()
            second_str = str(second).__repr__()
            if msg is None:
                msg = (f'Tested floats should match:'
                       f'\nFirst : {first_str} ({type(first)})'
                       f'\nSecond: {second_str} ({type(second)})')
            raise AssertionError(msg)


    def assertIn(self, member : object, container, msg : Optional[str] = None):
        if not member in container:
            member_str = str(member).__repr__()
            container_str = str(container).__repr__()
            if msg is None:
                msg = f'{member_str} not in {container_str}'
            raise AssertionError(msg)


    def assertIsInstance(self, obj : object, cls : type, msg : Optional[str] = None):
        if not isinstance(obj, cls):
            obj_str = str(obj).__repr__()
            cls_str = str(cls).__repr__()
            if msg is None:
                msg = f'{obj_str} not an instance of {cls_str}'
            raise AssertionError(msg)


    def assertTrue(self, expr : bool, msg : Optional[str] = None):
        if not expr:
            if msg is None:
                msg = f'Tested expression should be true'
            raise AssertionError(msg)


    def assertFalse(self, expr : bool, msg : Optional[str] = None):
        if expr:
            if msg is None:
                msg = f'Tested expression should be false'
            raise AssertionError(msg)




import logging
import os
import psutil

from holytools.logging import LoggerFactory
# -------------------------------------------

drive_logger = LoggerFactory.get_logger(name=__name__, include_location=False)
def log(msg : str, level : int = logging.INFO):
    drive_logger.log(msg=msg, level=level)
    print(drive_logger.name)


class PartitionInfo:
    @classmethod
    def from_resource_path(cls, resource_path : str):
        device_name = find_device_of_path(resource_path=resource_path)
        return cls(device_name=device_name)

    def __init__(self, device_name: str):
        self.mount_point : str = get_device_mount_path(device_name=device_name)
        if not self.mount_point:
            raise ResourceWarning(f'Could not find mount point for device {device_name}')

        self.device_name : str = device_name


    def print_free_space_info(self):
        warning_free_space = 250
        critical_free_space_in_GB = 100

        free_space_GB = int(round(self.get_free_space_in_GB(), 0))
        total_space_GB = int(round(self.get_total_space_in_GB(), 0))
        log(f'Free space on partition {self.device_name}: {free_space_GB}/{total_space_GB} GB')

        if free_space_GB < critical_free_space_in_GB:
            log(f'Warning: Almost no disk space remaining on {self.device_name}. Only {free_space_GB} GB left!', level=logging.CRITICAL)
        elif free_space_GB < warning_free_space:
            log(f'Warning: Disk space is running low on "{self.device_name}". Only {free_space_GB} GB left!', level=logging.WARNING)


    def get_free_space_in_GB(self) -> float:
        stats = psutil.disk_usage(self.mount_point)
        return stats.free / (1024 ** 3)

    def get_total_space_in_GB(self) -> float:
        stats = psutil.disk_usage(self.mount_point)
        return stats.total / (1024 ** 3)

def find_device_of_path(resource_path: str) -> str:
    absolute_path = os.path.abspath(resource_path)

    partitions = psutil.disk_partitions(all=True)
    partitions.sort(key=lambda x: len(x.mountpoint), reverse=True)

    device_name = None
    for partition in partitions:
        if absolute_path.startswith(partition.mountpoint):
            device_name = partition.device

    if not device_name:
        raise FileNotFoundError(f'Could not find device for path {resource_path}')

    return device_name


def get_device_mount_path(device_name : str) -> str:
    partitions = psutil.disk_partitions()
    mount_path = None
    for part in partitions:
        if part.device == device_name:
            mount_path = part.mountpoint

    if not mount_path:
        raise FileNotFoundError(f'Could not find mount point for device {device_name}')

    return mount_path


if __name__ == '__main__':
    test_partitions = psutil.disk_partitions()
    print(test_partitions)

    # test_part = PartitionInfo(device_name='/dev/nvme0n1p3')
    # log(test_part.mount_point)
    # test_part.print_free_space_info()
    print(find_device_of_path(resource_path='/'))



    # new_part = Partition.from_resource_path(resource_path='/media/daniel/STICKY1')
    # log(new_part.mount_point)
    # new_part.print_free_space_info()
    #
    # for p in partitions:
    #     print(p.mountpoint, psutil.disk_usage(p.mountpoint).percent)


from __future__ import annotations
import os.path
from typing import Optional

import secretstorage
from cryptoalgos.encrypt import AES
from cryptoalgos.hash import SHA

from holytools.configs.base import BaseConfigs
from holytools.logging import LogLevel

# ---------------------------------------------------------

class FileConfigs(BaseConfigs):
    def __init__(self, fpath : str = '~/.pyconfig', encrypted : bool = False):
        self._config_fpath: str = self._as_abspath(path=fpath)
        self.is_encrypted : bool = encrypted
        if self.is_encrypted and not self.keyring_available():
            raise ValueError('Encrypted configs can only be used with D-Bus Secret Service e.g. GNOME Keyring')

        if self.is_encrypted:
            self.aes : AES = AES()
            self.sha : SHA = SHA()
            self.masterpw_hash : bytes = self.retrieve_masterpw_hash()
        config_dirpath = os.path.dirname(self._config_fpath)
        os.makedirs(config_dirpath, exist_ok=True)
        super().__init__()

    @classmethod
    def credentials(cls) -> FileConfigs:
        try:
            fpath = os.environ[f'CRED_FPATH']
        except:
            raise ValueError(f'Environment variable \"CRED_FPATH\" not found. '
                             f'Please define file path of config file through environment variable \"CRED_FPATH\"')
        # print(f'fpath = {fpath}')
        if not os.path.isfile(fpath):
            raise ValueError(f'File at path \"{fpath}\" does not exist')

        return FileConfigs(encrypted=True, fpath=fpath)


    def _populate_map(self):
        if not os.path.isfile(self._config_fpath):
            self.log(msg=f'File {self._config_fpath} could not be found, configs are empty', level=LogLevel.WARNING)
            return self._map

        content = self.read()
        lines = content.split(f'\n')

        current_section = None
        non_empty_lines = [line for line in lines if line.strip()]
        for num, line in enumerate(non_empty_lines):
            parts = line.split(f' = ')
            if line.startswith('[') and line.endswith(']'):
                current_section = line[1:-1]
                self._map[current_section] = {}
            elif len(parts) == 2:
                self._read_line(parts=parts, current_section=current_section, num=num)
            else:
                raise ValueError(f'Line {num + 1} in config file is invalid: \"{line}\"')

    def _read_line(self, parts : (str, str), current_section : Optional[str], num : int):
        key, value = parts
        if ' ' in key:
            raise ValueError(f'Key must not contain whitespaces, got key \"{key}\" in line {num + 1}')
        if ' ' in value:
            raise ValueError(f'Value must not contain whitespaces, got value \"{value}\" in line {num + 1}')
        self._map[current_section][key] = value

    def _update_resource(self):
        general_section = {k:v for k,v in self._map[None].items()}
        other_sections = {k:v for k,v in self._map.items() if not k is None}

        config_content = ''
        for k,v in general_section.items():
            config_content += f'{k} = {v}\n'

        for section_name, section_map  in other_sections.items():
            config_content += f'\n[{section_name}]\n'
            for subkey, subval in section_map.items():
                config_content += f'{subkey} = {subval}\n'

        self.write(content=config_content)

    def read(self) -> str:
        if not os.path.isfile(self._config_fpath):
            raise ValueError(f'No file exists at configured config fpath {self._config_fpath}')

        with open(self._config_fpath, 'r') as f:
            content = f.read()
        if self.is_encrypted:
            content = self.aes.decrypt(content=content, key=self.masterpw_hash)

        return content

    def write(self, content : str):
        if self.is_encrypted:
            content = self.aes.encrypt(content=content, key=self.masterpw_hash)
        with open(self._config_fpath, 'w') as f:
            f.write(content)

    def retrieve_masterpw_hash(self) -> bytes:
        label_name ="masterpw"
        connection = secretstorage.dbus_init()
        collections = secretstorage.get_all_collections(connection)

        if not collections:
            raise KeyError("No keyring collections found.")

        master_pw_hash = None
        for collection in collections:
            items = collection.get_all_items()
            master_pw_matches = [i for i in items if i.get_label() == label_name]
            if master_pw_matches:
                master_pw_item = master_pw_matches[0]
                master_pw = master_pw_item.get_secret().decode('utf-8', errors='replace')
                master_pw_hash = self.sha.get_hash(txt=master_pw)

        connection.close()
        if master_pw_hash is None:
            raise KeyError(f'No password labeled "{label_name}" found in any keyring')

        return master_pw_hash

    @staticmethod
    def keyring_available() -> bool:
        try:
            connection = secretstorage.dbus_init()
            connection.close()
            return True
        except:
            return False


if __name__ == "__main__":
    master_pwd = FileConfigs(encrypted=True).retrieve_masterpw_hash()
    print(f'Master pw hash = {master_pwd}')
    creds = FileConfigs.credentials()

